
# coding: utf-8

# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>
# 
# <table style="width:95%;">
#   <tr style="background-color: #fff; font-family:'Helvetica'; font-size:120%">
#     <td style="padding-right:2em;">
#         <img src="https://crmda.dept.ku.edu/images/logoleft.png" alt="left side logo" style="width:60; height:60;">
#     </td>
#     <td style="width:60%; text-align:left; ">
#         <h1 style="text-align:center;">Python Data Science</h1>
#         <h3 style="text-align:center;">Working with Language: Modeling</h3>
#         <div style=" display: block; margin-top: 1em; margin-bottom: 1em; border-style: solid; border-width: 1px;"></div>
#         <span style="font-weight:bold;">Dr. Jonathan Lamb</span>, Faculty Fellow, CRMDA, <a class="author_email" href="mailto:#"> &lt;jonathanplamb@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Keywords:</span> web scraping, HTML parsing, table data
#     </td>
#     <td style="padding-left:2em;">
#     <img src="https://crmda.dept.ku.edu/images/logo-vert.png" alt="right side logo" style="width:60; height:60;">
#     </td>
# </tr>
# </table>
# 
# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>

# # More language work
# In the morning session, we developed a basic sense of how to use Python to work with texts. This afternoon, we are exploring some of the more advanced uses of Python. Although we won't get too far 'in the weeds' of each particular method or tool, we will discuss enough for you to get a sense of its use and features.
# 
# The scripts below are a mish-mash of new and borrowed scripts, courtesy of Ted Underwood, the UC-Berkeley DH lab, and Stefan Sinclair and Geoffrey Rockwell.

# # Document Similarity

# A common task when working with a larger corpus is to try to determine how documents relate to one another – how similar or different they are, or how they might cluster together. For instance, if we take the 18 texts from the NTLK Gutenberg corpus, are there any texts that group together? There are several authors with multiple texts (Shakespeare, Austen, Chesterton), do these cluster together?
# 
# More importantly, how do we determine similarity? For instance, we could plot the ratio of sentences per number of words and see if that's helpful in determining stylistic similarity.

# In[ ]:


get_ipython().magic('matplotlib inline')
import nltk

sentencelengths = {}
for fileid in nltk.corpus.gutenberg.fileids():
    sentencelengths[fileid] = len(nltk.corpus.gutenberg.sents(fileid)) / len(nltk.corpus.gutenberg.words(fileid))
    
nltk.FreqDist(sentencelengths).plot()


# ### Document Term Matrix
# Another way to consider document similarity is to consider and compare the frequency of terms in each document. The first step in doing this is to create what's called a document term matrix where frequencies are indicated for each document – it might look something like this:

# In[ ]:


import pandas as pd

# This cell creates a document term matrix of these sentences.

documents = [
    "I love jazz music and I love to read good books",
    "What good books have you read recently",
    "It's music to my ears when you say I love you"
]

allwords = set([word.lower() for word in nltk.word_tokenize(" ".join(documents)) if any ([c for c in word if c.isalpha()])])
allrawfreqs = [nltk.FreqDist(nltk.word_tokenize(document.lower())) for document in documents]
pd.DataFrame(allrawfreqs).fillna(0)


# We could create a similar table using relative frequencies instead, in other words, where each value would be relative to the total number of tokens in the document. That would be better for variable length documents, but an even more powerful (and commonly used) technique is to calculate a value that better represents the significance of a term within the corpus. Just as the relative frequency dampens counts based on document length, the TF-IDF value dampens relative frequencies based on the distribution of the term in the corpus.
# 
# ### TF-IDF
# We already have the "TF" part of the TF-IDF value, it's simply the term's raw or relative frequency in a given document. We need to multiply that by IDF, which is the inverse document frequency, which we can calculate like this:
# 
#     IDF = log of the total number of documents divided by the number of documents that contain the term
# 
# So, imagine we want to calculate the TF-IDF score of the word "boot" for Austen's _Emma_ within the Gutenberg corpus:
# 
# * 4 occurrences of "boot" in _Emma_
# * 16,1975 words in _Emma_
# * 6 documents that contain "boot" (at least once)
# * 18 documents in total
# 
#     TF = 4 / 161975 = 0.00002469516901
# 
#     IDF = log(18 /6) = 1.0986122886681098
# 
#     TF-IDF (for boot in _Emma_) = 0.00002469516901 \* 1.0986122886681098 = **0.00002713041615**
#     
# That seems like a lot of work for a single TF-IDF value for one term in one document. But good news! We don't need to calculate TF-IDF scores ourselves, there are several libraries that do that for us. In fact, we've already seen this with the gensim library in the topic modelling notebook, but here we're going to use [scikit](http://scikit-learn.org/stable/), a machine learning library for Python, and in particular with [TfidfVectorizer](http://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html).

# In[ ]:


from sklearn.feature_extraction.text import TfidfVectorizer
simpletm = TfidfVectorizer().fit_transform(documents) #takes a list of strings and tokenizes them for us
pd.DataFrame(simpletm.toarray())


# In[ ]:


names = nltk.corpus.gutenberg.fileids() # Keep track of file names (for labels)
texts = [nltk.corpus.gutenberg.raw(fileid) for fileid in names] # list of strings, one per document
documentTermMatrix = TfidfVectorizer().fit_transform(texts) #This performs the TF-IDF
documentTermMatrix


# Our document term matrix is stored in something called a sparse matrix, which is a useful and efficient way to store and process a very large table of data where several values are absent or not set. There are typically a lot of zero values in a sparse matrix (imagine a term that only occurs in one of the 18 documents – a true matrix would need to store 18 cells/values, but a sparse matrix can store just one.
# 
# This document term matrix is conceptually enormous. Depending on the arguments we provide to fit_transform() we could have a huge table of 18 (the number of documents) by the total number of unique words in all the documents (each term is a column and each column needs a value for each one of the documents).

# ### How to graph these relationships?
# 
# Even though we have a table of TF-IDF values for every term in every document, in order to make any use of it we need a way of comparing documents, in other words of determining the distance between the frequency values (columns) for any of the documents the rows. In effect, we want a way of expressing how different two rows (documents) are.
# 
# We can do this by converting our TF-IDF values from each document into geometric space. Imagine each term frequency (as TF-IDF) as a point on a cartesian graph and a value that expresses the aggregate of these points. Each document could thus be expressed as a vector, or a line with a certain magnitude (length) and angle from the origin (0, 0).
# 
# Because the length of the line (the total magnitude of the total TF-IDF values) would be sensitive to the length of the document, we can't simply compare the lines for each document, but we can compare the angle between the documents – that's the essence of cosine similarity.
# 
# * cosine distance between document 1 and document 2 (d1, d2)
# * cosine distance between document 2 and document 3 (d2, d3)
# * cosine distance between document 1 and document 3 (d1, d3)
# 
# We'll use scikit's [cosine_similarity](http://scikit-learn.org/stable/modules/metrics.html#cosine-similarity) function.

# In[ ]:


from sklearn.metrics.pairwise import cosine_similarity
distances = 1 - cosine_similarity(documentTermMatrix)


# In[ ]:


distances


# This creates a new matrix not of individual terms but of distance measures between documents. Think of a table of distances between cities on a map, if you present it in a table there will be duplication, and identity values for where cities meet.
# 
# <table><tr><td></td><td>Calgary</td><td>Montreal</td><td>Toronto</td><td>Vancouver</td></tr><tr><td>Calgary</td><td></td><td>3750</td><td>3450</td><td>1050</td></tr><tr><td>Montreal</td><td>3750</td><td></td><td>550</td><td>4800</td></tr><tr><td>Toronto</td><td>3450</td><td>550</td><td></td><td>4500</td></tr><tr><td>Vancouver</td><td>1050</td><td>4800</td><td>4500</td><td></td></tr></table>
# 
# The distance values that we get from cosine_similarity() on our documentTermMatrix from Gutenberg are similar, we have an 18 by 18 grid. But the grid is still difficult to read, we want a way of visualizing distances.

# ### Making distances visible
# 
# An 18 by 18 grid can be thought of as an 18 dimensional dataset. Coordinates on a cartesian graph are usually two dimensional, since we have x and y dimensions. So the question becomes: how do we convert a multidimensional grid into a two dimensional graph? Why, with [multidimensional scaling](https://en.wikipedia.org/wiki/Multidimensional_scaling) of course :). The math to accomplish this isn't as important for now as the purpose of multidimensional scaling to two dimensions, which is to represent as well as possible the distance that objects have in their original multiple dimensions.
# 
# This process leaves us with somethign we CAN graph--18 elements, each with an X and Y value that can be graphed in two-dimensional space.

# In[ ]:


from sklearn.manifold import MDS

# reduce n-dimensional distances matrix to a 2D matrix for x and y coordinates
mds = MDS(dissimilarity='precomputed', random_state=1)
positions = mds.fit_transform(distances)

positions

# this function leaves us with a list of 18 elements (one for each Gutenberg text)
# where each text has an x and y value for graphing.


# In[ ]:


# Now we just graph it!

import matplotlib.pyplot as plt
get_ipython().magic('matplotlib inline')

xvalues = positions[:, 0] # left column for x axis
yvalues = positions[: ,1] #right column for y axis

plt.figure(figsize=(10,10))
for x, y, name in zip(xvalues, yvalues, names):
    plt.scatter(x, y)
    plt.text(x, y, name.replace(".txt", ""))
plt.show()


# In[ ]:


# Another kind of visualization.

from scipy.cluster.hierarchy import ward, dendrogram
linkage_matrix = ward(distances)
dendrogram(linkage_matrix, labels=names, orientation='right');
plt.show()


# # Clustering and Topic Modeling

# The clustering section is modified from http://brandonrose.org/clustering . The scikit-learn documentation with examples is here: http://scikit-learn.org/stable/ . For further a explanation of topic modeling on the low-level see *Data Science from Scratch*: http://shop.oreilly.com/product/0636920033400.do .
# 
# Data you will need:
# 
# * The CMU book summaries dataset from here: http://www.cs.cmu.edu/~dbamman/booksummaries.html

# ### Preparing the data

# We're going to do some basic clustering and topic modelling of a large dataset of book summaries. These summaries were scraped from Wikipedia. **Question: how might this decision to scrape summaries from Wikipedia influence the dataset?**

# In[ ]:


import pandas as pd
import os

filepath = os.path.join('..', '..', 'data', 'booksummaries.txt')

df = pd.read_csv(filepath, sep="\t")


# In[ ]:


print(df.shape)


# In[ ]:


df.head()


# In[ ]:


lists = df.values.T.tolist()


# In[ ]:


print(lists[6][1])


# Now we can zip together the relevant data, we'll just be using the titles and summaries:

# In[ ]:


zipped_data = list(zip(lists[2],lists[6]))
small_data = zipped_data[:200] # just first 200

sums = [x[1] for x in small_data]
sums = [x.replace('"', "") for x in sums] # quotes are changed when tokenized and get caught
sums = [x.replace('...','') for x in sums] # elipses also get caught

titles = [x[0] for x in small_data]


# ## Clustering

# Clustering is an unsupervised machine learning method, as we are not telling the computer the groupings. Clustering is not the same as topic modeling, although clustering can yield topics. Clustering is a more restricted approach to grouping and visualizing data based on their similarity. If you only want to determine topics, a conventional LDA model will be more accurate as it allows for documents to be assigned to more than one topic. If you are looking for spatial relations and 1-1 assignment of documents to groupings, clustering will show this better.

# First we'll define functions in order to collect tokenized words and stemmed words:

# In[ ]:


import nltk
from nltk.stem.snowball import SnowballStemmer
from string import punctuation

#as these are summaries, we'll take out further stop words that dont' appear in enough summaries
more_stops = ['novel', 'book', 'books', 'story', 'narrator', 'narrative','character', 
              'chapters', 'chapter', 'tells', "'s"]

stemmer = SnowballStemmer("english")

def tokenize_only(text):
    tokens = [word.lower() for sent in nltk.sent_tokenize(text) for word in nltk.word_tokenize(sent)]
    filtered_tokens = [x for x in tokens if x not in punctuation 
                       and x not in more_stops] #word tokenizer cuts the possessives
    return filtered_tokens

def tokenize_and_stem(text):
    stems = [stemmer.stem(x) for x in tokenize_only(text)]
    return stems


# Now we collect these from our paragraphs.

# In[ ]:


totalvocab_stemmed = []
totalvocab_tokenized = []

for s in sums:
    totalvocab_stemmed.extend(tokenize_and_stem(s))
    totalvocab_tokenized.extend(tokenize_only(s))


# Our data frame will map tokenized words to stemmed words.

# In[ ]:


import pandas as pd

vocab_frame = pd.DataFrame({'words': totalvocab_tokenized}, index = totalvocab_stemmed)
print (vocab_frame.shape[0])
print (vocab_frame)


# In[ ]:


print(sums[0])


# We'll again use a tfidf matrix. To remind you: a tfidf takes into account the frequency of a word in the entire corpus, and offsets it based on its frequency among documents.

# In[ ]:


from sklearn.feature_extraction.text import TfidfVectorizer

#define vectorizer parameters, max_df is maximum occurence in docs of word, min_df is opposite
#use .8 max to eliminate more common words, lower .2 looking for unique but not proper nouns
#use inverse document frequency, give more weight to rare words
tfidf_vectorizer = TfidfVectorizer(max_df=0.8, max_features=20000,
                                 min_df=0.2, stop_words='english',
                                 use_idf=True, tokenizer=tokenize_and_stem, ngram_range=(1,5))

tfidf_matrix = tfidf_vectorizer.fit_transform(sums) #fit the vectorizer to the summaries

print(tfidf_matrix.shape)


# In[ ]:


print(tfidf_matrix)


# Then we need the words from the vector, these are essentially most influential words taking both document frequency and corpus frequency into account, we will eventually assign them to clusters.

# In[ ]:


terms = tfidf_vectorizer.get_feature_names()
print(terms)


# In order to plot our clusters in a 2D plane, we will again use the cosine similarity method.

# In[ ]:


from sklearn.metrics.pairwise import cosine_similarity
dist = 1 - cosine_similarity(tfidf_matrix) #creates high dimensional object
print (dist.shape)


# Now we'll start the actual clustering. The algorithm assigns each observation to the cluster whose mean yields the least within-cluster sum of squares, essentially the nearest mean. This iterates until the mean no longer changes.
# 
# But how do we know the number of clusters? This is a highly contentious matter. There are various methods (https://en.wikipedia.org/wiki/Determining_the_number_of_clusters_in_a_data_set), but we'll employ the silhouette method, which determines the similarity between documents within a cluster, and their relative similarity to documents in other clusters. We want the highest silhouette coefficient we can get.

# In[ ]:


from sklearn import metrics
from sklearn.cluster import KMeans

cnum = 6
km = KMeans(init='k-means++', n_clusters=cnum, n_init=10, random_state=10)
km.fit(tfidf_matrix)
clusters = km.labels_.tolist() #assigns each paragraph to the respective cluster


# We can now do a form of topic modelling by printing the words characterizing the clusters we made, the words are those closest to the centroid of the cluster, extracted from the vocab data frame, indexed by their position within the cluster:

# In[ ]:


#sort cluster centers by proximity to centroid, and grabs the index to iterate through below
order_centroids = km.cluster_centers_.argsort()[:, ::-1]

cents_words = [] #to collect words for chart legend

for i in range(cnum): #numer of clusters
    cent = []
    print("Cluster %i words:" % i, end='')
    
    for ind in order_centroids[i, :7]: #ind is index, replace 7 with n words per cluster, how many to choose from centroid
        a = vocab_frame.ix[terms[ind].split(' ')].values.tolist()[0][0] #indexing term in dataframe
        cent.append(a)
        
        print(' %a' % a, end=',')
        
    cents_words.append(cent)
    print ()


# To see which book summary was assigned to which cluster, we'll zip our titles and clusters together:

# In[ ]:


print(list(zip(titles,clusters)))


# ### Plot clusters

# Again, we'll use multidimensional scaling to create an X and Y axis for plotting.

# In[ ]:


from sklearn.manifold import MDS

# convert two components as we're plotting points in a two-dimensional plane
# "precomputed" because we provide a distance matrix
mds = MDS(n_components=2, dissimilarity="precomputed", random_state = 10)

pos = mds.fit_transform(dist)  # shape (n_components, n_samples), based on distances

xs, ys = pos[:, 0], pos[:, 1] #grabs x and y coordinates from pos (numpy array)


# Define colors and labels for plot:

# In[ ]:


import random

cluster_colors = {}
cluster_names ={}

cols = ["b", "g", "r", "c", "m","y"]

for i in range(cnum): # for each cluster
    #cluster_colors[i] = "#%06x" % random.randint(0, 0xFFFFFF) #random hexadecimal color
    cluster_colors[i] = cols[i]
    cluster_names[i] = ' '.join(cents_words[i][:4])


# Plot:

# In[ ]:


import matplotlib.pyplot as plt
get_ipython().magic('matplotlib inline')

#create data frame that has the result of the MDS plus the cluster numbers and titles
df = pd.DataFrame(dict(x=xs, y=ys, label=clusters, title=titles)) 

#group by cluster
groups = df.groupby('label')

# set up plot
fig, ax = plt.subplots(figsize=(17, 9)) # set size, subplots yields a tuple of figure and axes, hence the two assignments
ax.margins(0.15) # Optional, just adds 10% padding to the autoscaling

#iterate through groups to layer the plot
#note that I use the cluster_name and cluster_color dicts with the 'name' lookup to return the appropriate color/label
for name, group in groups:
    ax.plot(group.x, group.y, marker='o', linestyle='', ms=12, #marker size
            label=cluster_names[name], color=cluster_colors[name], 
            mec='none') #'marker edge color'
    
ax.legend(numpoints=1)  #show legend with only 1 point

#add label in x,y position with the label as the paragraph number
for i in range(len(df)):
    ax.text(df.ix[i]['x'], df.ix[i]['y'], df.ix[i]['title'], size=7)  

    
    
plt.show() #show the plot

#uncomment the below to save the plot if need be
#plt.savefig('clusters_small_noaxes.png', dpi=200)


# We can also cluster hierarchically, as before.

# In[ ]:


from scipy.cluster.hierarchy import ward, dendrogram

linkage_matrix = ward(dist) #define the linkage_matrix using ward clustering pre-computed distances

fig, ax = plt.subplots(figsize=(15, 40)) # set size
ax = dendrogram(linkage_matrix, orientation="right", labels=titles);

plt.tick_params(labelbottom='off')

plt.tight_layout() #show plot with tight layout

#uncomment below to save figure
# plt.savefig('ward_clusters.png', dpi=200) #save figure as ward_clusters


# ## Topic Modelling

# We can compare the above results with topic modeling results. There are two popular choices for models here: Latent Semantic Indexing (LSI) and Latent Dirichlet Allocation (LDA). LDA is a more complex process, and thus takes more resources and longer to run, but has higher accuracy. LSI is a much simpler process and can be run quite quickly.
# - LSI looks at words in a documents and its relationships to other words, with the important assumption that every word can only mean one thing. (cf. https://en.wikipedia.org/wiki/Latent_semantic_indexing)
# - LDA seeks to remedy this fault by allowing words to exist in multiple topics, first grouping them by topic, and each document is compared across each topic to determine the best fit. (cf. https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation)

# In[ ]:


# Try to run this cell. If you do and receive an error, 
# remove the hastag in front of the next cell and run it.

from gensim import corpora, models, similarities 


# In[ ]:


#!conda install gensim --y


# In[ ]:


from gensim import corpora, models, similarities 
from nltk.corpus import stopwords
from string import punctuation

sums_tm = [tokenize_only(x) for x in sums]
sums_tm = [[stemmer.stem(x) for x in i if x not in more_stops] for i in sums_tm]
sums_tm = [[x for x in i if x not in stopwords.words("english") and x not in punctuation] for i in sums_tm]


# We first create a dict of word IDs and their respective word frequency for all documents.

# In[ ]:


#create a Gensim dictionary from the texts
dictionary = corpora.Dictionary(sums_tm)

#remove extremes (similar to the min/max df step used when creating the tf-idf matrix)
#no_below is absolute # of docs, no_above is fraction of corpus
dictionary.filter_extremes(no_below=40, no_above=.70)


# The corpus we now create with doc2bow is a vector of all words (IDs from the dict), and frequency for each document.

# In[ ]:


#convert the dictionary to a bag of words corpus for reference
corpus = [dictionary.doc2bow(i) for i in sums_tm]


# ## LSI

# For LSI we first create a tfidf similar to above with clustering:

# In[ ]:


tfidf = models.TfidfModel(corpus)
corpus_tfidf = tfidf[corpus]


# We then build the model:

# In[ ]:


lsi = models.LsiModel(corpus_tfidf, id2word=dictionary, num_topics=6)
corpus_lsi = lsi[corpus_tfidf]
lsi.print_topics(6)


# Not so great, let's try LDA.

# ## LDA

# The LDA model can be built immediately:

# In[ ]:


#we run chunks of 15 books, and update after every 2 chunks, and make 10 passes
lda = models.LdaModel(corpus, num_topics=6, 
                            update_every=2,
                            id2word=dictionary, 
                            chunksize=15, 
                            passes=10)

lda.show_topics()


# In[ ]:


corpus_lda = lda[corpus_tfidf]
for i,doc in enumerate(corpus_lda): # both bow->tfidf and tfidf->lsi transformations are actually executed here, on the fly
    print(titles[i],doc)
    print()


# For more with gensim, see the tutorials here: https://radimrehurek.com/gensim/tutorial.html

# ## Word2Vec, what the heck?

# Word embeddings are the first successful attempt to move away from the "bag of words" model of language. Instead of looking at word frequencies, and vocabulary usage, word embeddings aim to retain syntactic information. Generally, a word2vec model *will not* remove stopwords or punctuation, because they are vital to the model itself.
# 
# word2vec simply changes a tokenized sentence into a vector of numbers, with each unique token being its own number.
# 
# e.g.:
# 
# ~~~
# [["I", "like", "coffee", "."], ["I", "like", "my", "coffee", "without", "sugar", "."]]
# ~~~
# 
# is tranformed to:
# 
# ~~~
# [[43, 75, 435, 98], [43, 75, 10, 435, 31, 217, 98]]
# ~~~
# 
# Notice, the "I"s, the "likes", the "coffees", and the "."s, all have the same assignment.
# 
# The model is created by taking these numbers, and creating a high dimensional vector by mapping every word to its surrounding, creating a sort of "cloud" of words, where words used in a similar syntactic, and often semantic, fashion, will cluster closer together.
# 
# One of the drawbacks of word2vec is the volume of data necessary for a decent analysis. So we will read in a copy of the King James Bible and hope it will provide enough data, it then needs to be broken into sentences and tokenized:

# In[ ]:


import os
import nltk

filepath = os.path.join('..', '..', 'data', 'King_James_Bible.txt')

with open(filepath, "r") as f:
    bible = f.read()

bible = nltk.sent_tokenize(bible)
bible = [nltk.word_tokenize(s) for s in bible]


# Now we can actually train the model on the language of the Bible:

# In[ ]:


bible_words = 0
for sent in range(len(bible)):
    bible_words += len(bible[sent])
print(bible_words)


# In[ ]:


from gensim.models import Word2Vec
model = Word2Vec(bible, size=300, window=5, min_count=5, workers=4)
model.train(bible, total_words=bible_words, epochs=model.iter)


# In[ ]:


model.wv.most_similar('man')


# In[ ]:


model.wv.most_similar('woman')


# We can even create little equations, so what would be a:
# 
# KING + WOMAN - MAN = ?

# In[ ]:


model.wv.most_similar(positive=['king', 'woman'], negative=['man'])


# In[ ]:


print(model.wv.similarity('man', 'woman'))
print(model.wv.similarity('God', 'man'))


# Play around a bit with the word2vec model and see what words seem to have interesting 'neighborhoods.' Feel free to view the [word2vec documentation](https://radimrehurek.com/gensim/models/word2vec.html) for more ideas about how this powerful modelling works.
