
# coding: utf-8

# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>
# 
# <table style="width:95%;">
#   <tr style="background-color: #fff; font-family:'Helvetica'; font-size:120%">
#     <td style="padding-right:2em;">
#         <img src="https://crmda.dept.ku.edu/images/logoleft.png" alt="left side logo" style="width:60; height:60;">
#     </td>
#     <td style="width:60%; text-align:left; ">
#         <h1 style="text-align:center;">Python Data Science</h1>
#         <h3 style="text-align:center;">Working with Data: If There's Time</h3>
#         <div style=" display: block; margin-top: 1em; margin-bottom: 1em; border-style: solid; border-width: 1px;"></div>
#         <span style="font-weight:bold;">Dr. Jonathan Lamb</span>, Faculty Fellow, CRMDA, <a class="author_email" href="mailto:#"> &lt;jonathanplamb@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Keywords:</span> Dunnings, Log-likelihood, Twitter
#     </td>
#     <td style="padding-left:2em;">
#     <img src="https://crmda.dept.ku.edu/images/logo-vert.png" alt="right side logo" style="width:60; height:60;">
#     </td>
# </tr>
# </table>
# 
# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>

# # The geometry of meaning
# 
# We're going to explore some basic forms of text analysis, using David Robinson's dataset of tweets made from the account of Donald J. Trump. The basic premise is this: in 2016, Robinson and others noted that sometimes Trump tweeted from an iPhone, and sometimes from an Android. It quickly became clear that Trump used one device and his aides used another.
# 
# To begin, let's import some modules we're going to need later, and also read in the Trump data.

# In[ ]:


import os, csv, math
import pandas as pd
import numpy as np

from collections import Counter
from sklearn.decomposition import TruncatedSVD
from sklearn.preprocessing import Normalizer
      
relativepath = os.path.join('..', '..', 'data', 'trump.csv')
trump = pd.read_csv(relativepath)


# ## Different ways of identifying "distinctive" words
# 
# In this section we'll explore Dunning's log-likelihood, and also think about the strengths and weaknesses of "distinctive" words as evidence.
# 
# First let's glance at the Trump dataset.

# In[ ]:


trump.head()


# #### Basic functions
# 
# For a lot of the work we do today, we're going to want to construct dictionaries that hold the frequencies of words in different categories: Trump-iphone or Trump-android. To do this we'll need to break text into words, count the words in each text, and then add up the counts by category.
# 
# Let's define some functions that do this. (Note that these are hacky versions of what we learned with ```nltk```.)

# In[ ]:


def tokenize(astring):
    ''' Breaks a string into words, and counts them.
    Designed so it strips punctuation and lowercases everything,
    but doesn't separate hashtags and at-signs.
    '''
    wordcounts = Counter()
    # create a counter to hold the counts
    
    tokens = astring.split()
    for t in tokens:
        word = t.strip(',.!?:;-—()<>[]/"\'').lower()
        wordcounts[word] += 1
        
    return wordcounts


# In[ ]:


def addcounters(counter2add, countersum):
    ''' Adds all the counts in counter2add to countersum.
    Because Counters(like dictionaries) are mutable, it
    doesn't need to return anything.
    '''
    
    for key, value in counter2add.items():
        countersum[key] += value


# In[ ]:


def create_vocab(seq_of_strings, n):
    ''' Given a sequence of text snippets, this function
    returns the n most common words. We'll use this to
    create a limited 'vocabulary'.
    '''
    vocab = Counter()
    for astring in seq_of_strings:
        counts = tokenize(astring)
        addcounters(counts, vocab)
    topn = [x[0] for x in vocab.most_common(n)]
    return topn


# In[ ]:


# Let's test the vocabulary function.
vocab = create_vocab(trump['text'], 4000)
vocab[0:10]
trump.shape


# #### A few more basic functions
# 
# Now that we have a vocabulary, we're going to want to divide our texts into categories, create Counters summing the word frequencies in those categories, and then compare the two Counters to find words that are overrepresented in one category relative to the other.
# 
# There are several ways we could define "overrepresented." We'll use Robinson's simple log-odds measure, as well as Dunning's log-likelihood.

# In[ ]:


def logodds(countsA, countsB, word):
    ''' Straightforward.
    '''
    
    odds = (countsA[word] + 1) / (countsB[word] + 1)
    
    # Why do we add 1 on both sides? Two reasons. The hacky one is 
    # that otherwise we'll get a division-by-zero error whenever
    # word isn't present in countsB. The more principled reason
    # is that this technique (called Laplacian smoothing) tends
    # to reduce the dramatic disproportion likely to be found in
    # very rare words.
    
    return math.log(odds)


# In[ ]:


def signed_dunnings(countsA, totalA, countsB, totalB, word):
    ''' Less straightforward. This function calculates a signed (+1 / -1)
    version of Dunning's log likelihood. Intuitively, this is a number 
    that gets larger as the frequency of the word in our two corpora
    diverges from its EXPECTED frequency -- i.e., the frequency it would
    have if it were equally distributed over both. But it also tends to get
    larger as the raw frequency of the word increases.
    
    Note that this function requires two additional arguments:
    the total number of words in A and B. We could calculate that inside
    the function, but it's faster to calculate it just once, outside the function.
    
    Also note: the strict definition of Dunnings has no 'sign': it gets bigger
    whether a word is overrepresented in A or B. I've edited that so that Dunnings
    is positive if overrepresented in A, and negative if overrepresented in B.
    '''
    if word not in countsA and word not in countsB:
        return 0
    
    # the raw frequencies of this word in our two corpora
    # still doing a little Laplacian smoothing here
    a = countsA[word] + 0.1
    b = countsB[word] + 0.1
    
    # now let's calculate the expected number of times this
    # word would occur in both if the frequency were constant
    # across both
    overallfreq = (a + b) / (totalA + totalB)
    expectedA = totalA * overallfreq
    expectedB = totalB * overallfreq
    
    # and now the Dunning's formula
    dunning = 2 * ((a * math.log(a / expectedA)) + (b * math.log(b / expectedB)))
    
    if a < expectedA:
        return -dunning
    else:   
        return dunning


# In[ ]:



# a set of common words is often useful
stopwords = {'a', 'an', 'are', 'and', 'but', 'or', 'that', 'this', 'so', 
             'all', 'at', 'if', 'in', 'i', 'is', 'was', 'by', 'of', 'to', 
             'the', 'be', 'you', 'were'}

# finally, one more function: given a list of tuples like
testlist = [(10, 'ten'), (2000, 'two thousand'), (0, 'zero'), (-1, 'neg one'), (8, 'eight')]
# we're going to want to sort them and print the top n and bottom n

def headandtail(tuplelist, n):
    tuplelist.sort(reverse = True)
    print("TOP VALUES:")
    for i in range(n):
        print(tuplelist[i][1], tuplelist[i][0])
    
    print()
    print("BOTTOM VALUES:")
    lastindex = len(tuplelist) - 1
    for i in range(lastindex, lastindex - n, -1):
        print(tuplelist[i][1], tuplelist[i][0])
        
headandtail(testlist, 2)
    


# ## Exercise 1: Is Dunning's a better measure than logodds for Trump's tweets?
# 
# In other words, which of the two predictive functions provides a more accurate rendering of Trump's Twitter language? Let's put together all the functions we've written to answer that question.
# 
# See what happens if you do (or don't) remove stopwords and tweets that begin with a quote.
#                                                    
# Then we'll try Dunning's log likelihood. Does that seem to be a better (more revealing) measure of overrepresentation? How would we decide?

# In[ ]:


# Start by creating a vocabulary for words in the Trump tweets.
# Put it in a variable called 'vocab'.

# Remember the function create_vocab takes two arguments:
# (seq_of_strings, n)
# We can afford to include all the words, so set n for 5000.

vocab = create_vocab(trump['text'], 5000)

# An optional step: removing stopwords
vocab = list(set(vocab) - stopwords)


# In[ ]:


numrows = len(trump.index)
print(numrows)


# In[ ]:


# Create counters for the android and iphone corpora.

android = Counter()
iphone = Counter()

androidwords = 0
iphonewords = 0

# Figure out how many rows are in the Trump DataFrame
# and put that number in a variable like 'numrows.'
# Then iterate through the 'text' column of the data frame.

# for each text cell, get a Counter with words counts for that cell
# then add those counts either to iphone or android.

# When you get around to running Dunning's, you'll need to
# create variables that hold the total count of *all words*
# in iphone and android.

for i in range(numrows):
    counts = tokenize(trump['text'][i])
    if 'iphone' in trump['statusSource'][i]:
        addcounters(counts, iphone)
        iphonewords += sum(counts.values())
    elif 'android' in trump['statusSource'][i]:
        addcounters(counts, android)
        androidwords += sum(counts.values())

print(androidwords)
print(iphonewords)
        


# In[ ]:


# Create an empty list to hold pairs of (overrepresentation_measure, word)
# Then iterate through your vocabulary. For each word, measure 
# overrepresentation using either logodds or signed_dunnings.
# Create a tuple, (overrepresentation_measure, word)
# and append it to the empty list you created.

wordmeasures = []

for word in vocab:
    wordmeasures.append((logodds(iphone, android, word), word))


# In[ ]:


# Finally use the headandtail function to display the top 25 and bottom 25
# words in your tuplelist.

headandtail(wordmeasures, 25)


# In[ ]:


dunningswordmeasures = []

for word in vocab:
    dunningswordmeasures.append((signed_dunnings(iphone, iphonewords, android, androidwords, word), word))


# In[ ]:


headandtail(dunningswordmeasures, 25)


# ## Using corpora to create a "meaning space."
# 
# Contrasting two corpora can be revealing, but sometimes we want to think about the relations between individual texts or words. To do that, we often represent them as vectors in a multi-dimensional space.
# 
# The simplest way to do this is to create a document-term matrix, something we encountered on Day 2. Here's a function that does that. It requires a pre-defined vocabulary (list of words) as well as a list (or numpy vector) of texts.

# In[ ]:


def doc_term_matrix(vocab, textvector):
    ''' Transform the textvector into a document-term matrix
    with one column for each word in vocab.
    '''
    
    n = len(textvector)
    vocabset = set(vocab)
    # making a set so we can check membership quickly;
    # it's much faster in a set than in a list
    
    termdictionary = dict()
    for word in vocab:
        termdictionary[word] = np.zeros(n)
    for i, text in enumerate(textvector):
        counts = tokenize(text)
        for word, count in counts.items():
            if word in vocabset:
                termdictionary[word][i] += count
    
    dtmatrix = pd.DataFrame(termdictionary, columns = vocab)
    return dtmatrix

# A nice arcane trick to perform on a document-term matrix
# is to squash it into a smaller number of dimensions. This
# often reveals relationships between words that don't
# necessarily, literally occur together. The technique is called
# Latent Semantic Analysis.

def lsa_matrix(dtmatrix, vocab, number_of_dimensions):
    lsa = TruncatedSVD(number_of_dimensions, algorithm = 'arpack')
    dtm_lsa = lsa.fit_transform(dtmatrix)
    dtm_lsa = Normalizer(copy=False).fit_transform(dtm_lsa)
    lsamatrix = pd.DataFrame(lsa.components_, columns = vocab)
    
    return lsamatrix

def cosine_similarity(vector1, vector2):
    dot_product = np.sum(vector1 * vector2)
    # we assume these are numpy vectors and can be
    # multiplied elementwise
    
    magnitude = math.sqrt(sum([val**2 for val in vector1])) * math.sqrt(sum([val**2 for val in vector2]))
    if not magnitude:
        return 0
    else:
        return dot_product/magnitude


# ## Exercise 2: Finding words that are close in "meaning space."
# 
# As we did on Day 2, we'll measure semantic similarity as the cosine similarity between vectors defined by a word's distribution across documents.
# 
# Let's try this in the space defined by Trump tweets.

# In[ ]:


# Let's start by getting a vocabulary,
# and a doc-term matrix, as well as a
# squashed (lsa) version of that matrix.

vocab = create_vocab(trump['text'], 5000)
dtm = doc_term_matrix(vocab, trump['text'])
lsa = lsa_matrix(dtm, vocab, 25)
dtm.head()


# Next, let's write a function that receives an LSA matrix and a single word, and then returns words close in 'meaning space' to the single word.

# In[ ]:


def find_matches(amatrix, word):
    columnA = amatrix[word]
    
    vocab = amatrix.columns.values
    tuplelist = []
    for w in vocab:
        columnB = amatrix[w]
        cosinesim = cosine_similarity(columnA, columnB)
        tuplelist.append((cosinesim, w))
    
    headandtail(tuplelist, 10)


# In[ ]:


user_word = input('word? ')
find_matches(lsa, user_word) 

