## Python Data Science
### Working with Internet Data: Retrieval
### Charles Redmon <redmon@ku.edu>

# # Introduction
# 
# ## Structure of the web
# 
# - Web pages are text files
# - In the olden days, it was fairly easy to write them by hand because nobody expected them to be very fancy
# - In 1992, the "here is the simplest web page ever" document was literally no more than this
# 
# ```html
# <html>
#     <head>
#     <title>Simplest Web Page Ever</title>
#     <link rev=made href="mailto:pauljohn@ku.edu">
#     </head>
#     <body>
#     This is <i>italics</i>
#     <p>
#     This is <b>bold faced</b>
#     <p>
#     A hyperlink is <a href="http://crmda.ku.edu">The Fabolous CRMDA Website</a>
#     <p>
#     </body>
# </html>
# ```

# - It would display in a browser like this:
# 
# <img src="./simplest-webpage.png" alt="simple webpage">

# - We've been teaching HTML Web programming for beginners for more than 20 years with an example that is no more complicated than that (The [original simplest page ever](http://pj.freefaculty.org/misc_support_docs/htmlhelp/samplePage.html) 
# - The images don't show anymore because a server was shut down, but otherwise, it is as lovely as ever!
# 
# ### The core idea
# - Web browsers retrieve pages and display them to us
# - That process becomes more-and-more complicated because standards for writing pages become more complicated over time
# - But, for the most part, this works

# ## Tags within `<>`
# 
# - A web page's code can be reviewed in a Web browser
# - Currently in Firefox, one can click Tools -> Web Developer -> Page Source (or the shortcut Control-U) to find out what the Web browser retrieved from a website
# - Here are things that you will notice:
# 
#     1. Some tags are "paired". There is an opener, `<b>` and a closer `</b>`. Notice that the main tags, such as `html`, `head`, `body`, are all entered in pairs. Web links, which are inside `<a href="....">` and `</a>` are similarly paired.  If these tags are not correctly entered as pairs, then the Web browser may not be able to make sense of it.
#     2. Some tags do not need to be paired, such as the new paragraph `<p>`. However, in the newest style of Web code, paragraphs also are encased in `<p>` and `</p>`.
#     3. The tags are case-insensitive. In the olden days, we always used capital letters. Now the small letters are the norm.
#     4. The part of the page inside `<head>` and `</head>` does not show in the browser. But it can be important. It can include Javascript code to enhance the beauty of the page. It is a place where a "style sheet" can be included.

# ### Tags are also useful for other items
# 
# Numbered lists:
# 
# ```html
#      <ol>
#          <li>First</li>
#          <li>Second</li>
#      </ol>
# ```
# Which would appear as:  
# 
# 1. First   
# 2. Second
#   
#   
# Bullet (unnumbered) lists:
# 
# ```html
# <ul>
#     <li>First</li>
#     <li>Second</li>
# </ul>
# ```
# Which would appear as:
# 
# <ul>
#      <li>First</li>
#      <li>Second</li>
# </ul>                  

# ## HTML includes a language for tables
# 
# - In web scraping, our purpose is often to open up a page's text and find a particular element, often a table, and harvest the data
# - Because that is our goal, we need to explain how tables are formatted in HTML
# - Here is the idea:
#     - A table begins with `<table>` and ends with `</table>`
#     - A new row in the table begins with `<tr>` and a column's worth of information is typed between `<td>` and `</td>` tags
#     - In case you want a first row to appear as a header, the column tags would be `<th>`. 
# 
# - Here is code for a simple 3 row table:
# 
# ```html
# A table with 3 columns
# <table>
#     <tr><th>City</th><th>Happy</th><th>Sad</th></tr>
#     <tr><td>Denver</td><td>90</td><td>10</td></tr>
#     <tr><td>Kansas City</td><td>80</td><td>20</td></tr>
#     <tr><td>Oakland</td><td>5</td><td>95</td></tr>
# </table>
# ```
# 
# - That will display like so:
# 
#     A table with 3 columns
# 
# <table>
# <tr><th>City</th><th>Happy</th><th>Sad</th></tr>
# <tr><td>Denver</td><td>90</td><td>10</td></tr>
# <tr><td>Kansas City</td><td>80</td><td>20</td></tr>
# <tr><td>Seattle</td><td>5</td><td>95</td></tr>
# </table>
# 
# - This one looks pretty good
# - Its display is controlled by a "style sheet" that is applied to all tables within this document
# - We did not exert any effort to control its placement, the alternating colors, and so on
# - At the current state of development in web programming, it would not be too difficult to fiddle those details, but for the most part we are discouraged to avoid it
# - We preserve the overall "look and feel" of the document by being cooperative

# ## Style Sheets and Class Identifiers
# 
# - In the olden days, we would write formatting statements into the `<table>` tag itself
# - As you see on the manuals for web writers hosted by [mozilla.org](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/table), detailed formatting within tables is  **deprecated**
# - Instead, one should put formatting statements into a "style sheet" and then apply the style sheet to the table
# 
# 
# - It turns out that the styling code used with tables will sometimes help us sort through Web documents
# 
# ### The best case scenario: An `id` tag.
# 
# - An id tag is a unique element, it can be used by only one object in a document
# - Supposing a style sheet defines a style element called "redtable", then when they use it the table code will begin with 
# 
# ```
# <table id="redtable">
# ```
# 
# - *If* the Web page's author uses a particular labeled style for this one table that we *really* want, then it is very easy to find it within the page code
# 
# ### The next best case: A class assigned to tables of a particular format
# 
# - If there's no id tag in use, too bad
# - We might be forced to inspect the page looking at lots of tables
# - Perhaps we can narrow down the possibilities if the Web page author declared the tables with classes
# - For example, if the page has 50 tables, maybe we will be lucky and there are only 2 or 3 tables declared with a particular class:
# 
# ```
# <table class="greentable">
# ```
# 
# - The class designator may not be unique, so we might have to inspect several tables, but it can help us to narrow down the possibilities
# 
# ### The next best case: Study the idiosyncracies of the Web author
# 
# - It often happens that Web programmers have their own opinions about how tables ought to be presented
# - Some will ignore the style sheet altogether, and they will literally write style commands into the table call, either by the old-style format or the new style tags
# - The newest standard HTML5 does not support attributes declared in the old style, but they may still be in use
# 
# ```html
# <table align="left" bgcolor="#FFFFCC" border="3" cellpadding="5" cellspacing=2">
# ...
# ```
# 
# - More modern programmers would probably have the settings in a style sheet, but many of us who create just one table of a particular sort are not too proud to write a style specifications into the `<table>` declaration
# 
# ```
# <table style="background-color=#FFFFFF; border: solid #000 3px; ...
# ```
# 
# - If we are trying to find a particular table within a larger page, we might get lucky and find some of this "hand coded" style markup at the beginning of a particular table
# - This is almost as informative as an `id` tag, in the sense that we have a more or less mechanical way to figure out what we need to keep
# 
# 
# #### \*\*We will return to more of this structure in later sessions, but for now this is sufficient background to forge ahead

# # The web scraping process
# 
# - At its simplest, *web scraping* involves two procedures:
#   1. Retrieving (accessing and downloading) web pages in the form of `.html` files
#   2. Parsing the (largely) HTML code to get the data in a usable format
#   <br><br>
# - This lesson focuses primarily on the *retrieval* process
# - Some parsing procedures will be introduced to facilitate the transition from static web page retrieval to the more dynamic procedure of *web crawling*
# 
# 
# - Before jumping into web data retrieval methods, we must first discuss briefly the issue of ethics accessing and downloading Internet data
# 
# ## Ethics
# 
# - Given the public nature of the data displayed on the web, it may not be immediately obvious that there are ethics involved in the use of this data
# - See Thelwall & Stuart (2006) for an in-depth review
# - A number of considerations are recommended to ensure that your web scraping/crawling procedure is consistent with the rights and expectations of the web maintainer whose website(s) we are extracting data from
# - The following outline is based on Thelwall & Stuart (2006), where a more thorough consideration of the issues involved may be found (see also Markham et al., 2012 for the latest report from the Ethics Working Committee of the Association of Internet Researchers)<br>
#     - *Denial of Service*
#         - Websites have limited network bandwidth and repeated requests (particularly to a website not designed for that volume) can overload the server<br><br>
#     - *Cost*
#         - Websites pay hosting providers for a specific bandwidth allottment, and must pay additional fees whenever this bandwidth is exceeded
#         - Often these fees are tied to the amount of extra bandwidth used (e.g., 2 cents per Mb), so an unconstrained crawler can rack up fees for the website owner<br><br>       
#     - *Privacy*
#         - Even if web data is not secured (i.e., does not require a specific user login) and may be assumed to be in the public domain, it is worth considering whether the sensitivity of the data extracted from a given page/site<br><br>
#     - *Copyright*    
#         - Making copies of websites could potentially infringe on copyright, so web maintainers are generally instructed to indicate which aspects of the site should not be copied
#         - This information is provided in the `robots.txt` file, which is described below
# 
# 
# ### `robots.txt`
# - At the top level of each site there should be a file named `robots.txt` that lists data/pages which are and are not permitted for scraping/crawling
# - In a later example, we will look at statutes of the Kansas judiciary, which are hosted on the site `ksrevisor.org` 
# - The following is the content of the `robots.txt` file provided by that site:
# 
# ```
# # robots.txt for http://www.ksrevisor.org
# 
# User-agent: *
# Disallow: /images/
# Disallow: /cgi-bin/
# Disallow: /err/
# Disallow: /css/
# Disallow: /sjtest/
# Disallow: /scripts/
# ```
# 
# - This file tells us we are allowed to scrape all content except that whose parent folder is `ksrevisor.org/images/`, `ksrevisor.org/cgi-bin/`, etc.
# 
# ## Retrieving HTML files
# 
# - If you're new to web scraping, this may be the first time you will access a web site without the help of a browser
# - But in principle, there is little difference between requesting a page through the URL menu in a browser, and requesting a page by submitting a URL string to a function in Python
# - In Python, the module used for this procedure is the appropriately named `requests` module

# In[1]:


import requests


# - Next we define the a variable (for ease of re-use) with the text of the website we plan to visit
# - We will be using Ukrainian election data as the initial example here
# 
# ### 2014 Ukrainian Election
# 
# - District-level voting data from the election is provided at the following site:

# In[2]:


uaURL = "http://www.cvk.gov.ua/pls/vnd2014/wp039"


# - Here's a preview of what this webpage looks like: <br>
# <img src="./ukraineWebpage.png" alt="ukrainian election webpage">

# - To grab this page and use it in our Python session we use the `get` method in the `requests` module
# - But for the sake of sparing the Ukrainian government's servers, we won't run this code here
# - Below we provide the code we originally ran to get the data
# 
# ```python
# uaPage = requests.get(uaURL)
# ```
# 
# - The first argument is simply the URL string
# - There is a second argument we can specify for more complex queries, which is a dictionary of parameters (shown in the url as "www.../webpage**?key=value**" for `params={"key":"value"}`) that may be necessary to access particular content
# - These parameters are not necessary for now, but we'll return to them later when we start working with APIs
# 
#   
# 
# - After retrieving the page, we need to check status of the request (did we successfully access the page from the web?)
# - `uaPage.status_code` will just provide a number (like 200, 404, etc.), the meaning of which may not be obvious
# - The requests package provides a dictionary of codes (which is too long to print here), but the most important one to check is `requests.codes.ok`, which is equal to 200, and should be the code returned by the page retrieved via `requests.get` if that retrieval was successful
# 
# ```python
# uaPage.status_code == requests.codes.ok
# True
# ```
# 
# - Next we write this data to a local file so that we can work with that file instead of re-downloading it every time we work with the script
# - You should see this file as `ukraineElection2014.html` in the `data` directory
# - First we check the methods available for accessing data stored in `uaPage` (excluding the internal methods)
# 
# ```python
# [m for m in dir(uaPage) if m[0] != "_"]
# ['apparent_encoding', 'close', 'connection', 'content', 'cookies', 'elapsed', 'encoding', 
#  'headers', 'history', 'is_permanent_redirect', 'is_redirect', 'iter_content', 'iter_lines', 
#  'json', 'links', 'next', 'ok', 'raise_for_status', 'raw', 'reason', 'request', 'status_code', 
#  'text', 'url']
# ```
# 
# - The one we want is `content`, which returns a byte string that when written to a local HTML file will retain the critical data from the online web page
# - Note: some of the styling of the page will naturally disappear as we have not downloaded the supporting CSS files
# 
# 
# ```python
# import os
# ddir = "../../data"
# filename = "ukraineElection2014.html"
# with open(os.path.join(ddir, filename), "wb") as file:
#     file.write(uaPage.content)
# ```

# - So now we have our data stored locally
# - We can import it and work with it as we would an HTML file we pulled directly from the web

# In[3]:


import os
import codecs
ddir = "../../data"
filename = "ukraineElection2014.html"
uaFile = codecs.open(os.path.join(ddir, filename), 'r', 'windows-1251') 
uaText = uaFile.read()
uaText


# - Most of this should be review, but how did we get to know that the encoding is `windows-1251`?
# - When you request the page you can get information about its encoding through the `requests.encoding` method
# - We'll get more into the details of HTML parsing in the next session, but for now we'll explore just a few core aspects of working with HTML data in Python, as they'll be necessary when we introduce web crawling in the next section.
# 
# 
# - For parsing HTML, we use the `BeautifulSoup` function in the `bs4` package
# - `BeautifulSoup()` extracts the structure contained in the HTML tags in the plain text string `uaText`
# - But first, we use the `.prettify` method to print the contents in a cleaner way than what we saw above for the raw text

# In[4]:


from bs4 import BeautifulSoup as bs
uaHTML = bs(uaText, "lxml") # lxml is a specific parser (others may need to be tried if output is not satisfactory)
uaHTML.prettify


# - There's other information in the `BeautifulSoup` object, which again we find out by applying the `dir()` function to `uaHTML` 
# - (We could also read the documentation, of course)

# In[5]:


[m for m in dir(uaHTML) if m[0] != "_"]


# In[6]:


uaHTML.title


# In[7]:


uaHTML.head


# In[8]:


uaHTML.find_all("li")


# In[9]:


uaHTML.find_all("a")


# In[10]:


[a.text for a in uaHTML.find_all("a")]


# In[11]:


uaHTML.a.attrs


# In[12]:


[a["href"] for a in uaHTML.find_all("a")]


# ## 'Crawling'
# 
# - The term "web scraping" refers generally to a **static** process
# - Web "crawling" is a **dynamic** procedure, where we use the webpage we scraped to find new URLs to follow
# - To demonstrate this process, we look at judicial statutes in Kansas, which are provided on the web in a series of separate pages that are inconvenient to download manually
# 
# ### Kansas Mining Statutes
# 
# - Before we begin, here's a preview of the page we're scraping:
# 
# <img src="./ks-mining-statutes.png" alt="Kansas mining statutes webpage">
# 
# - As you can see, we have a list of lists of statutes, each with its own link to a separate page containing information on that statute
# - Just as with the Ukraine example, we first request the page we need and save it locally
# 
# ```python
# kssURL = "http://www.ksrevisor.org/statutes/ksa_ch49.html"
# kssPage = requests.get(kssURL)
# filename = "ksMiningStatutes.html"
# with open(os.path.join(ddir, filename), "wb") as file:
#     file.write(kssPage.content)
# ```

# In[13]:


filename = "ksMiningStatutes.html"
ksFile = codecs.open(os.path.join(ddir, filename), 'r') 
ksText = ksFile.read()
ksHTML = bs(ksText, "lxml")
ksHTML.prettify


# - Note that all the critical information we need for each statute is stored on a separate page
# - The links to these pages are arranged in a **list** environment
# - So broadly, what we want to do to grab all that data and store it locally for use later is:
# 
#     1. Grab each link from the list
#     2. Download the webpage corresponding to each link via `requests.get` (as before)
#     3. Write the HTML that is currently in Python memory to a local file
# 
# 
# - We will run through most of this procedure, only replacing the request and download part with a simpler print statement to avoid hitting the server with too many requests
# - Because we don't want everyone in the class hitting the server with requests, we executed the following beforehand
# - First, we create a directory to store the files, called `ks-mining-statutes`
# - We then create a variable in Python referencing this location

# In[14]:


sdir = os.path.join(ddir, "ks-mining-statutes")


# - Looking through the HTML with the browser's developer tools, we find that our list of statutes (with corresponding links) is in a `<div>` environment with a unique class `ksa`
# - We link to this structure with the following code:

# In[15]:


ksa = ksHTML.find("div", attrs={"class":"ksa"})


# - Our list is then a `<ul>` environment (which is actually a list of lists embedded in another `<ul>` environment)
# - We first select the top-level `<ul>` with `.find("ul")`:

# In[16]:


ksaList = ksa.find("ul")


# - Then we create a list of our lists (the list of subcategories that can be expanded on the website to show the actual statute links)

# In[17]:


statSecs = ksaList.find_all("ul")
statSecs


# - Now, for a given list, if we loop through all the items in that list we can get the links we need and follow them to the webpages we want to download
# - Let's first prove we can iterate through the list by simply printing the links we would need to follow 
# - Later we'll show the code we used to actually download the pages 

# In[18]:


for section in statSecs:
    
    statutes = section.find_all("li")
    
    for s in statutes:
    
        linkEnv = s.find("a")
        link = linkEnv["href"]
        url = "http://www.ksrevisor.org" + link
        print(url)


# - Now, to get to the code that actually requests the pages and downloads them, it's merely a matter of replacing `print()` with a function (or sequence of code) that does what we did before to request pages and write out `.html` files
# - Here is the function we used:
# 
# ```python
# def downloadPage(x):
#     page = requests.get(x)
#     if page.status_code == requests.codes.ok:
#         with open(os.path.join(sdir, os.path.basename(x)), "wb") as file:
#             file.write(page.content)
# ```
# 
# - With this function we can then just replace `print(url)` with `downloadPage(url)` in the previous code block
# - The HTML files downloaded in this manner should be in your directory under `data/ks-mining-statutes/`
# - In the next session, we will work with the files in that folder, as well as our Ukrainian election file, to go into greater detail on parsing HTML to access and organize specific pieces of data/content

# # References
# 
# - Markham, A., Buchanan, E., & the AoIR Ethics Working Committee. (2012). Ethical decision-making and Internet research: Version 2.0. *Association of Internet Researchers*.
# 
# - Thelwall, M., & Stuart, D. (2006). Web crawling ethics revisited: Cost, privacy, and denial of service. *Journal of the Association for Information Science and Technology, 57*(13), 1771-1779.
