## Python Data Science
### Working with Internet Data: Parsing
### Charles Redmon <redmon@ku.edu>

# # Parsing hierarchical data structures
# 
# - We have already seen a bit of HTML parsing with `BeautifulSoup`
# - In this session we focus a bit more on the HTML structure, and how we can take advantage of it to extract the data we need
# - Here is a tree diagram of some of the major components of the Ukrainian Election web page:

# <img src="htmlTree.png" alt="XML tree">

# - Stripping out some of the irrelevant code, we have the following HTML that maps on to this diagram:

# ```html
# <html dir="ltr" lang="ru-RU">
#     <head>
#         <title>Центральна виборча комісія України</title>
#     </head>
#     <body>
#         <table class="t1">
#             <tr>
#                 <td><img src="/images/check_tick_old1.gif" align="absmiddle" border=0></td>
#                 <td>Відомості про підрахунок голосів виборців в межах одномандатних виборчих округів</td>
#             </tr>
#         </table>
#         <table class="t2">
#             <tr>
#                 <td align="center" width="10%">ОВО №11</td>
#                 <td align="center" width="50%">Домбровський Олександр Георгійович...</td>
#                 <td align="center" width="10%">51.37</td>
#             </tr>
#             <tr>
#                 <td align="center" width="10%">ОВО №12</td>
#                 <td align="center" width="50%">Порошенко Олексій Петрович...</td>
#                 <td align="center" width="10%">64.04</td>
#             </tr>
#         </table>
#     </body>
# </html>
# ```

# - The important thing to understand is that HTML is **hierarchical**
# - To access various elements we have to move down the tree
# - We'll illustrate this process by further parsing our two example data sets: Ukrainian election data and Kansas mining statutes

# ## 2014 Ukrainian election (cont.)
# 
# - First let's load in the data

# In[1]:


# packages for reading from local files in different formats
import os
import codecs

# package for parsing HTML
from bs4 import BeautifulSoup as bs

# data path and file name
ddir = "../../data"
filename = "ukraineElection2014.html"

# open connection to file
uaFile = codecs.open(os.path.join(ddir, filename), 'r', 'windows-1251') 

# read in file
uaText = uaFile.read()

# convert text to Beautiful Soup structure
uaHTML = bs(uaText, "lxml")


# - Looking at the web page, we can see that the data we want is provided in a six-column table
# - If you go to the URL, you can use your browser's developer tools (typically envoked with F12) to inspect that table

# <img src="develTools.png" alt="developer tools demo">

# - We see that it is an HTML `<table>` structure
# - Before we used the `find_all` method to extract elements
# - From here on out we will use `select`, because of its greater flexibility in the long run
# - With `select`, information on element attributes is added in square brackets as below

# In[2]:


t2_tabs = uaHTML.select("table[class=t2]")


# - Did we get what we wanted?
# - To check whether there are other tables with the "t2" class, we just check the length of `tab_t2`

# In[3]:


len(t2_tabs)


# - Looking back at the HTML with the Developer Tools view in the browser, we can see that it is the second of the two "t2" tables that we want
# - The first is a header of some sort

# In[4]:


tab = t2_tabs[1]
tab


# - We can further traverse the elements of this table by selecting child elements (possibly with various attributes)
# - The `>` syntax specifies that `<a>` is a direct child of `<td>`

# In[5]:


distLinks = tab.select("td[class=td3] > a")
distLinks


# - To access attributes within this list, we use the `get` method in a list comprehension structure

# In[6]:


[l.get("href") for l in distLinks]


# - Now you can see how we might use this traversing procedure for each row in the table, but there are easier and better ways
# - In the `pandas` package there is a method called `read_html`, that takes a table string and converts it to a list of data frames
# - If we only have one data frame we will need to extract the first element of the list to get out of the list class

# In[7]:


import pandas as pd
df = pd.read_html(str(tab))
df = df[0]
df.head()


# - You can see here that we're not out of the woods yet
# - There are a few rows that are just header/summary nonsense
# - We can subset the data to start at row 3, but first we should save row 0 as the header

# In[8]:


df.columns = df.iloc[0]
df.head()


# - Now subset:

# In[9]:


df1 = df[3:]
df1.head(10)


# - This is starting to look better, but note that row 11 has some missings
# - If you look back at the web page, you'll see that these are intermediate headers of some sort
# - We want to remove them
# - First, get the rows where we find NaN in the 3rd column:

# In[10]:


col3 = df1.iloc[:, 2].apply(str) # convert each element to a string
col3[:10]


# In[11]:


dropIdxs = [i for i,x in enumerate(col3) if x == "nan"]
dropIdxs


# - Now drop these rows from the data frame:

# In[12]:


df2 = df1.drop(df1.index[dropIdxs])
df2.head(10)


# - Now the final thing we want to do is fix the 4th column, which has spaces in the numbers

# In[13]:


df2.iloc[:, 3] = [str(i).replace(" ", "") for i in df2.iloc[:, 3]]
df2.head()


# - Let's write this out as a CSV for use later

# In[14]:


wdir = "../../workingdata" 
df2.to_csv(os.path.join(wdir, "ukraineElection2014_data.csv"), index=False)


# ## Kansas Mining Statutes (cont.)
# 
# - Earlier we saw how to crawl through the links to separate Kansas mining statutes and save the files locally (this could have been done more efficiently with the `select` method, but we'll leave any cleaning up of the code as an exercise)
# - Now we want to parse each HTML file and store the data in a single tabular structure
# - First we load one file to see how the general parsing procedure will work; later we will incorporate this all into a loop over the whole directory

# In[15]:


# set path to statute directory
sdir = os.path.join(ddir, "ks-mining-statutes")

# list files in directory
files = os.listdir(sdir)
files


# In[16]:


# choose first one
testFile = codecs.open(os.path.join(sdir, files[0]), 'r') 

# read in file
testText = testFile.read()

# convert text to Beautiful Soup structure
testHTML = bs(testText, "lxml")


# - Extract data, which is in `<div id="print"`

# In[17]:


testData = testHTML.select("div[id=print]")
testData = testData[0]
testData


# - We will extract 4 pieces of information to include in the table:
#     1. Number (`<span class="stat_number">`)
#     2. Short description (`<span class="stat_caption">`)
#     3. Long description (`<p class="ksa_stat">`)
#     4. History (`<span class="history">`)

# In[18]:


import re
statNum = testData.select("span[class=stat_number]")  # get statute number
statNum = statNum[0].text  # pull out only the text from the <span> element
statNum = statNum.strip()  # remove leading and trailing whitespace
statNum = re.sub("\.", "", statNum)  # strip off final period
statNum


# In[19]:


shortDesc = testData.select("span[class=stat_caption]")
shortDesc = shortDesc[0].text
shortDesc = shortDesc.strip()
shortDesc


# In[20]:


desc = testData.select("p[class=ksa_stat]")[0]
desc


# In[21]:


descText = desc.find_all(text=True, recursive=False)
descText


# In[22]:


longDesc = descText[-1]
longDesc = longDesc.strip()
longDesc


# In[23]:


hist = testData.select("p[class=ksa_stat_hist]")[0]
hist = hist.find_all(text=True, recursive=False)
hist = hist[-1]
hist = hist.strip()
hist


# - Now let's roll this all into a loop to extract these values and append them to a data frame
# - First we create functions to extract each element:

# In[24]:


def getStatNum(x):
    y = x.select("span[class=stat_number]")
    if len(y) == 0:
        return("")
    else:
        y = y[0].text
        y = y.strip()
        y = re.sub("\.", "", y)
        return(y)

def getShortDesc(x):
    y = x.select("span[class=stat_caption]")
    if len(y) == 0:
        return("")
    else:
        y = y[0].text
        y = y.strip()
        return(y)

def getLongDesc(x):
    y = x.select("p[class=ksa_stat]")
    if len(y) == 0:
        return("")
    else:
        y = y[0].find_all(text=True, recursive=False)
        y = y[-1]
        y = y.strip()
        return(y)

def getHist(x):
    y = x.select("p[class=ksa_stat_hist]")
    if len(y) == 0:
        return("")
    else:
        y = y[0].find_all(text=True, recursive=False)
        y = y[-1]
        y = y.strip()
        return(y)


# - Now we loop through the list of files, each time parsing them and appending to a data frame

# In[25]:


# initialize list to accept data frames
datList = []

# initialize index (for tracking data frame rows)
i = 0

# loop through files
for f in files:
    
    # open connection, read file, and convert to BeautifulSoup object
    fileCon = codecs.open(os.path.join(sdir, f), 'r')
    fileTxt = fileCon.read()
    fileHTML = bs(fileTxt, "lxml")
    
    # get element containing all relevant data (in <div id="print">)
    dEnv = fileHTML.select("div[id=print]")
    dEnv = dEnv[0]
    
    # create data dictionary (to become columns in data frame), 
    #   applying the relevant function for each column
    dat0 = {"Statute": getStatNum(dEnv), 
            "Short_Description": getShortDesc(dEnv),
            "Long_Description": getLongDesc(dEnv),
            "History": getHist(dEnv)}
    
    # convert data dictionary to data frame object (index tracks file number)
    dat = pd.DataFrame(data=dat0, index = [i])
    
    # append to data list
    datList.append(dat)
    
    # update index
    i += 1

# concatenate all data frames into a single data frame
ksData = pd.concat(datList)

# reorder columns
ksData = ksData[["Statute", "Short_Description", "Long_Description", "History"]]

ksData.head()


# In[26]:


ksData.to_csv(os.path.join(wdir, "ksMiningStatutes_data.csv"), index=False)


# # Moving Forward
# 
# - We've successfully navigated HTML structure from multiple websites, and now have data that is much more manageable to work with
# - But there are still some formatting issues to handle before our data is ready for analysis
# - Much of this work will require the use of regular expressions, which we introduce in the next session
