## Python Data Science
### Working with Internet Data: Regular Expressions
### Charles Redmon <redmon@ku.edu>

# # Introduction to regular expressions
# 
# ## What you may have heard
# 
# The general reputation not so good. This is a common trope:
# 
# <img src="./perl_problems.png" alt="XKCD: Perl Problems" >
# 
# Our perspective is a bit more positive.
# 
# <img src="./regular_expressions.png" alt="XKCD: Regular Expressions" >
# 
# - Before Python, Perl was the go-to programming language for text processing
# - Much of the standard regular expression syntax is based on how it was implemented in Perl

# ## What are regular expressions?
# 
# - In essence, they are a concise syntax for pattern matching
#     - For example, `(?<=\.)\d*` matches ANY number of digits ONLY when preceded by a decimal point
#     - So this regular expression would produce the following matches: 3.<u>14159</u>, 2.<u>7</u>, etc.
# - They replace what would otherwise be a much more complex series of control statements (loops and conditionals)

# ## Common uses
# 
# - Checking if a string matches a specific pattern (usually for program control purposes)
# - Finding a string or set of strings within a text or set of texts
# - Replacing one string pattern with another
# - Splitting a string into substrings based on a pattern

# # Regular expressions in Python
# 
# - Regular expression functions in Python are included in the `re` module

# In[1]:


import re


# - We can list out all of `re`'s methods with the `dir` function, ignoring internal functions as before

# In[2]:


[m for m in dir(re) if m[0] != "_"]


#  - We can also just focus on the lowercase ones, which tend to be those we want anyway

# In[3]:


[m for m in dir(re) if m[0] != "_" and m[0].islower()]


# - Now we will do something that is quite *meta*
# - We can use a regular expression to list the methods in the regular expression module

# In[4]:


[m for m in dir(re) if re.match("^[^_A-Z].*", m)]


# - We will cover many of these functions in the examples to follow, but the primary workhorse in all regular expression procedures will be `re.match` (the one we just used)
# - `re.match` is the function that parses and accepts substrings according to regular expression syntax
# 
# - Let's demonstrate a few simple examples before tackling regular expressions more thoroughly

# ## Examples
# 
# ### Splitting a large text string into sentences

# In[5]:


crmda = "Interested in Learning More? Join Our Email List. Receive news, scholarly updates, workgroup reminders and presentation/event announcements. Another major objective of CRMDA is to cultivate and nurture methodological expertise across the University. This occurs in several ways: We sponsor the KU Stats Camp each summer. Registration open now! Click here for details. We organize Saturday Software Workshops that overview leading computer software. We prepare guides on the use of various programs and offer them online. We sponsor Workgroups for group collaboration and networking. We also monitor statistical course offerings throughout the campus so that students (and their advisors) can make well informed decisions about their opportunities."
crmda


# In[6]:


re.split("[!\.\?]\s+", crmda)


# In[7]:


# retain punctuation
re.split("(?<=[!\.\?])\s+", crmda)


# ### Extracting area codes from a string of phone numbers

# In[8]:


phoneNumbers = "(101) 482-3782; (332) 514-8152; (332) 888-1949; (696) 515-4728; (591) 231-3306; (827) 167-1653; (439) 867-8741; (827) 662-6706; (332) 414-6707; (101) 731-5461"
areaCodes = re.findall("\(\d{3}\)", phoneNumbers)
print(set(areaCodes))


# - We can remove the parentheses in a number of ways
# - The method below uses a bit more complicated syntax, which we'll introduce later

# In[9]:


areaCodes = re.findall("(?<=\()\d{3}(?=\))", phoneNumbers)
print(set(areaCodes))


# - Here is another way:

# In[10]:


areaCodes = re.findall("\((\d{3})\)", phoneNumbers)
print(set(areaCodes))


# - And here is yet another way:

# In[11]:


# first find the codes with parentheses
areaCodes = re.findall("\(\d{3}\)", phoneNumbers)
# next, replace parentheses with empty strings (i.e., delete them)
areaCodes = [re.sub("[\(\)]", "", code) for code in areaCodes]
print(set(areaCodes))


# ### Selecting columns in a data frame based on name structure
# 
# - Sometimes we have data that looks like this:

# In[12]:


import pandas as pd
import numpy as np
df = pd.DataFrame(np.random.randint(low=-7, high=7, size=(5, 10)),
                  columns=['ID', 'CLASS', 'MAT1', 'MAT1S', 'MAT2', 'MAT2S', 'VRB1', 'VRB1S', 'VRB2', 'VRB2S'])
df.head()


# - Ignore the numbers (they're just dummies to sketch out the data frame structure)
# - We may want just the standardized variables (`MAT1S`, `MAT2S`, etc.) to use in a model or plot or something
# - How do we select only these columns without changing the original data structure or hand entering column numbers?
# - A regex can be defined to match just these columns:

# In[13]:


df[[col for col in df.columns if re.match('.+\dS', col)]]


# ### Cleaning up file names
# 
# - In case you haven't heard already, **spaces in file names are BAD**
# - Other punctuation, besides hyphens, underscores, and periods, is also considered bad
# - We can clean them up pretty easily with regular expressions

# In[14]:


files = ["my awesome  data.xlsx", "I didn't know what to name this.ppt", "final revision 78!!!!.docx"]
## replace spaces with underscores
files = [re.sub("\s+", "_", f) for f in files]
## replace unwanted characters with empty string
files = [re.sub("[^\w\-\.]", "", f) for f in files]
files


# # A brief tour of regular expression syntax
# 
# ## Plain characters

# - We start with something trivial but important: **Most characters, when placed in a regular expression, are matched *as-is***
# - For example:

# In[15]:


re.findall("fish", "one fish, two fish, red fish, blue fish")


# - This goes for Unicode characters as well:

# In[16]:


# find the 'fish' (मछली, machhlii) in "one fish, two fish..." in Hindi
re.findall("मछली", "एक मछली, दो मछली, लाल मछली, नीला मछली")


# - Most characters are matched directly in regular expressions, though there are exceptions
# - The following characters / character sequences have special functions in the regular expression syntax:
# 
# Symbol      | Meaning 
# ------------|---------
# `\`         | Escape special char or start a sequence
# `.`         | Match any char except newline
# `^`         | Match start of the string
# `$`         | Match end of the string
# `[]`        | Enclose a set of matchable chars
# `R` $|$ `S` | Match either regex R or regex S
# `()`        | Create capture group, & indicate precedence
# `-`         | A range (inside `[]`), e.g. `a-c` matches a, b or c
# `^`         | Negate the set only if it is the 1st char
# `*`         | 0 or more
# `+`         | 1 or more
# `?`         | 0 or 1
# `{m}`       | Exactly m repetitions
# `{m,n}`     | From m (default 0) to n (default infinity)
# `\b`        | Match empty string at word (\w+) boundary
# `\B`        | Match empty string **not** at word boundary
# `\d`        | Digit
# `\D`        | Non-digit
# `\s`        | Whitespace [\t\n\r\f\v]
# `\S`        | Non-whitespace
# `\w`        | Alphanumeric: [0-9a-zA-Z_]
# `\W`        | Non-alphanumeric
# `\n`        | ASCII Linefeed
# `\t`        | ASCII Tab
# 
#  <p style="text-align:right"><em>adapted from Jonathan Hartley's <a href="https://github.com/tartley/python-regex-cheatsheet/"> Python regex cheatsheet</a></em></p> 
# 
# - To use them explicitly, they must be 'escaped'; i.e., written with a backslash in front (e.g., `\? \\.` )
# - Quiz: How do we explicitly match a single backslash?

# ## Optional matching

# ### Optional matching a preceding character: `?`
# 
# - Putting a question mark after a character makes that character optional in the match
# - To take a famous example:

# In[17]:


re.findall('colou?r', "Americans write 'color'. The British write 'colour'. But 'colr' is a misspelling.")


# ### Logical disjunction: `[ ]`, `|`
# - To specify a match on either among a set of characters, the square brackets `[ ]` are used  

# In[18]:


re.findall("sw[iau]m", "Swim, swam, swum.")


# - Multiple regular expressions (including simple strings of characters) can be put in an OR relation with the pipe symbol, `|`

# In[19]:


re.findall("sw[iau]m|d[iu]g", "Swim, swam, swum; dig, dag?, dug.")


# ## Quantifiers
# 
# - `.` : Match any character

# In[20]:


re.findall(".", "u5cvO/$7oQG4a")


# - `+` : Match 1 or more repetitions of a character or pattern
#     - Often this is combined with other special characters, such as the alphanumeric `\w`

# In[21]:


re.findall("\w+(?=\.)", "file1.txt, http://file2.html, file_3.R")


# - `*` : Match 0 or more repetitions of a character or pattern

# In[22]:


datVars = ['x1', 'log_x1', 'x2', 'log_x2', 'x3', 'log_x3']
[var for var in datVars if re.match('log.*', var)]


# ## Lookarounds
# 
# - Many times you want to match a particular pattern only when it occurs in a certain context
# - We saw this above with the file name matching, and with the sentence splitting
# - Logically we have four options: matching what follows (**positive lookahead**), matching what doesn't follow (**negative lookahead**), matching what precedes (**positive lookbehind**), matching what doesn't precede (**negative lookbehind**)
# 
# **Positive lookahead** : `(?=...)`

# In[23]:


re.findall("[\w\s]+(?=:)", 
           """for x in xlist:
                  print(x)""")


# **Negative lookahead** : `(?!...)`

# In[24]:


syntaxErrors = re.findall("[^;]+(?!;)\n",
"""
int s1len, s2len;
s1len = strlen(s1)
s2len = strlen(s2);
int column[s1len+1]
""")

for l in syntaxErrors:
    print(l)


# **Positive lookbehind** : `(?<=...)`

# In[25]:


re.findall("(?<=https://)[\w\.\-]+", "https://www.secure-site.com, http://www.unsecure-site.com")


# **Negative lookbehind** : `(?<!...)`

# In[26]:


re.findall("(?<!ftp:/)/\w+\.\w+\.?\w*", "https://www.site1.org, http://site2.org, ftp://user.site3.org")


# ## Testing regular expressions
# 
# - There are many online and offline tools for testing regular expressions
# - One fairly simple one that is specific to the regex syntax in Python is <a href="https://www.pythex.org">Pythex.org</a>:
# 
# <img src="./pythex.png" alt="Pythex: Python regex tester">

# # 2014 Ukrainian election (cont.)
# 
# - Recall we left off with the following table of election results scraped from the web:

# In[27]:


import os
wdir = "../../workingdata"
uaDF = pd.read_csv(os.path.join(wdir, "ukraineElection2014_data.csv"))
uaDF.head()


# - For ease of future manipulation, let's change the header names to something simple (and in a language I can understand)

# In[28]:


uaDF.columns = ["District", "Candidate", "VotePct", "Votes", "ProcessedPct", "Result"]


# In[29]:


uaDF.head()


# - There are three things I'd like to have different in this data set:
#     1. Instead of the text "ОВО №11" for the district number, I'd rather just have the number
#     2. The `Candidate` column actually has two pieces of information: candidate and affiliation. It would be helpful to have separate columns here
#     3. The `Result` column also contains a date which could be useful to have on its own in a date-formatted column

# ## Extracting district number
# 
# - Converting the `District` column into a column of integers is relatively straightforward
# - We just need to match only the digits and return them to the same column
# - First, let's test that the matching works:

# In[30]:


[re.sub("\D+", "", dist) for dist in uaDF.District]


# - Great! What did we do?
# - `\D` matches any character that IS NOT a digit
# - `+` specifies that we want to match 1 or more occurrences of the preceding (non-digit) character
# - `re.sub` takes all matches of 1 or more non-digit characters and substitutes them with and empty string `""`
# - Now let's save this result back to the `District` column

# In[31]:


uaDF["District"] = [re.sub("\D+", "", dist) for dist in uaDF.District]
uaDF.District.dtype  # check type


# In[32]:


uaDF["District"] = uaDF.District.astype("i")  # change to integer
uaDF.District.dtype


# In[33]:


uaDF.head()


# ## Creating a candidate affiliation column
# 
# - The second task, to create a column for candidate affiliation, involves splitting each string in the `Candidate` column on the comma, and saving the part that comes after in a second column
# - First let's check the column to see if there are any complications to deal with in the text:

# In[34]:


for cand in uaDF["Candidate"]:
    print(cand)


# - Fortunately, there appears to be only a single comma (distinguishing candidate from affiliation) in each cell
# - So if we simply split the text on this comma (handling surrounding spacing), we should get our two elements: candidate and affiliation
# - First we test to see if this is indeed the case
# - What do we get if we apply the `re.split` on each element

# In[35]:


[re.split("\s*,\s*", cand) for cand in uaDF.Candidate]


# - First, the pattern `\s*,\s*` specifies that we want to split on a comma preceded/followed by 0 or more whitespace characters
# - The result is we get a list of 2-element lists, the first being the candidate name and the second the candidate's affiliation
# - So now to save these as separate lists, we'll run the same code as above, but first selecting the first element, and then selecting the second, each time saving to a new variable

# In[36]:


candidates = [re.split('\s*,\s*', cand)[0] for cand in uaDF.Candidate]
affiliations = [re.split('\s*,\s*', cand)[1] for cand in uaDF.Candidate]
uaDF['Candidate'] = candidates
uaDF['Affiliation'] = affiliations
uaDF.head()


# - Now just clean up the column ordering so the table is easier to read:

# In[37]:


uaDF = uaDF[['District', 'Candidate', 'Affiliation', 'VotePct', 'Votes', 'ProcessedPct', 'Result']]
uaDF.head()


# ## Creating a date column
# 
# - Finally, we want to extract the date from the `Result` column
# - This will be a bit trickier but not too bad
# - First we test whether we can match just the date:

# In[38]:


[re.search("[\d\.]+", res).group(0) for res in uaDF.Result]


# - This looks correct
# - Now we want to save it to a new column and convert it to the proper date format

# In[39]:


uaDF["Date"] = [re.search("[\d\.]+", res).group(0) for res in uaDF.Result]
uaDF["Date"] = pd.to_datetime(uaDF.Date, dayfirst=True)
uaDF.head()


# - Finally, to remove redundancies, let's strip off the date from the `Result` column

# In[40]:


uaDF['Result'] = [re.sub('\([\d\.]+\)', '', res) for res in uaDF.Result]
uaDF.head()


# - Great! This table looks much cleaner and more manageable for analysis now

# # Kansas Mining Statutes (cont.)
# 
# - Here's where we left off last time:

# In[41]:


ksDF = pd.read_csv(os.path.join(wdir, "ksMiningStatutes_data.csv"))
ksDF.head()


# - I'm not going to worry about the complexity in the description for now
# - The main goal here will be to transform the `History` column into the following more useful columns:
#     - `Enacted`: The date the statute was enacted
#     - `Revised`: The latest revision date (if applicable)
#     - `Published`: The published date on the statute
#     - `Repealed`: The repeal date (if applicable)
#     
#     
# - First, we'll create our four columns and populate them with NaNs

# In[42]:


import numpy as np
ksDF['Enacted'] = np.nan
ksDF['Revised'] = np.nan
ksDF['Published'] = np.nan
ksDF['Repealed'] = np.nan
ksDF.head()


# - Next, before we do our text processing on the `History` column, we need to convert it to character format (`.astype("U")` specifies unicode characters, which is standard for strings in Python 3)

# In[43]:


ksDF['History'] = ksDF.History.astype("U")


# - Now for the main event!
# - We'll loop through each row of the `History` column and apply several regular expressions to check for different patterns corresponding to each column

# In[44]:


## initialize loop with enumerate() to track indices
for i, h in enumerate(ksDF.History):
    
    ## Regex 1: find years that are preceded by "L." but not "Repealed, L."
    ## (?:                        nested lookbehind (for reasons I will not get into)
    ##    (?<=                    positive lookbehind (match if preceded by)
    ##        (?<!                negative lookbehind (don't match if preceded by)
    ##            Repealed,\s     match "Repealed, "
    ##        )                   close group (i.e., don't match if preceded by "Repealed, ")
    ##         L\.\s              match "L. "
    ##    )                       close group (i.e., match if preceded by "L. " but not "Repealed, L. ")
    ## )                          close group
    ##  \d{4}                     match exactly 4 digits (for year)
    yrs = re.findall("(?:(?<=(?<!Repealed,\s)L\.\s))\d{4}", h)
    
    ## if result of regex match has any entries, 'Enacted' column receives the first entry
    if len(yrs) > 0:
        ksDF.loc[i, 'Enacted'] = yrs[0]
    ## if result has more than 1 entry, there must be a published date after the enactment date
    ##   - use last year in this case as published date
    if len(yrs) > 1:
        ksDF.loc[i, 'Published'] = yrs[-1]
    
    ## Regex 2: find years preceded by "R.S. " (for revised statute)
    ## (?<=            positive lookbehind (match if preceded by)
    ##     R\.S\.\s      match "R.S. "
    ## )               close group (i.e., match if preceded by "R.S. ")
    ##  \d{4}          match exactly 4 digits (for year)
    yrRevised0 = re.findall("(?<=R\.S\.\s)\d{4}", h)
    
    ## Regex 3: find years preceded by "R. S. " 
    ##   This regex is the same as above but with an extra space
    ##   Unfortunately this is necessary because lookbehind must be fixed-width (no optional matching)
    yrRevised1 = re.findall("(?<=R\.\sS\.\s)\d{4}", h)
    
    ## if regex 2 returns value, use it (assumes R.S. and R. S. are not both in same line)
    if yrRevised0:
        ksDF.loc[i, 'Revised'] = yrRevised0[-1]
    ## if regex 3 returns value, use it (assumes R.S. and R. S. are not both in same line)
    elif yrRevised1:
        ksDF.loc[i, 'Revised'] = yrRevised1[-1]
    
    ## Regex 4: find years preceded by "Repealed, L. "
    ## (?<=                    positive lookbehind (match if preceded by)
    ##     Repealed,\sL\.\s    match "Repealed, L. "
    ## )                       close group (i.e., match if preceded by "Repealed, L. ")
    ##  \d{4}                  match exactly 4 digits (for year)
    yrRepealed = re.findall("(?<=Repealed,\sL\.\s)\d{4}", h)
    
    ## if regex 4 returns value, fill in 'Repealed' column with value
    if yrRepealed:
        ksDF.loc[i, 'Repealed'] = yrRepealed[0]


# In[45]:


ksDF.head()


# - Now our History data is much more usable for analysis

# # Moving ahead
# 
# - We hope to have demonstrated a range of powerful and unique uses of regular expressions in data processing
# - To get fully comfortable with regular expressions you would probably need a few weeks
# - But since these situations are not uncommon when dealing with web data, we expect when you encounter them in the 
