## Python Data Science
### Working with Internet Data: APIs
### Charles Redmon <redmon@ku.edu>

# ### Modules required for this session

# In[1]:


import os
import requests
import re
import json
import xml.etree.ElementTree as et
import pandas as pd
from matplotlib import pyplot as plt


# ### Optional modules
# 
# ```python
# import twitter
# import nasa
# ```

# # What is an API?

# - Whenever anyone says they're scraping data from Twitter, or Facebook, or some other major social media platform, odds are they're not doing what we just did in the past few sessions
# 
# - Why is that? Largely for two reasons:
#      1. APIs deliver data in a format that is often more convenient for programmers
#      2. Large companies like Twitter and Facebook are protective of their data, and APIs allow them to share it but only under their terms 
# 
# 
# - Broadly, an API is a way of communicating with various components of a piece of software (this includes websites)
# - Since data are one such component, we can use an API to ask a website for some of its data
# - The biggest thing to know about web scraping using an API is that **APIs are idiosyncratic**
# - If you want to use Twitter's API, or Facebook's, or Google Maps, etc., you have to read the API documentation on their respective websites (e.g., https://developer.twitter.com/docs/, https://developers.facebook.com/docs/, https://developers.google.com/maps/documentation/)
# - We will get to queries soon enough (we'll introduce them from within Python), but first let's look at some common data structures used by APIs

# # API data structures
# 
# ## XML
# 
# - As we've been working with HTML a bit, XML should feel familiar
# - Here is an example of an XML data structure returned by the OpenWeatherMap API:
# 
# ```xml
# <current>
#     <city id="4274277" name="Lawrence">
#         <coord lon="-95.24" lat="38.97"/>
#         <country>US</country>
#         <sun rise="2018-04-24T11:30:46" set="2018-04-25T01:07:43"/>
#     </city>
#     <temperature value="297.929" min="297.929" max="297.929" unit="kelvin"/>
#     <humidity value="27" unit="%"/>
#     <pressure value="996.4" unit="hPa"/>
#     <wind>
#         <speed value="3.04" name="Light breeze"/>
#         <gusts/>
#         <direction value="14" code="NNE" name="North-northeast"/>
#     </wind>
#     <clouds value="0" name="clear sky"/>
#     <visibility/>
#     <precipitation mode="no"/>
#     <weather number="800" value="clear sky" icon="01d"/>
#     <lastupdate value="2018-04-24T18:43:30"/>
# </current>
# ```
# 
# ## JSON
# 
# - JSON stands for **J**ava**S**cript **O**bject **N**otation, and is one of the more common data structures returned by APIs
# - A JSON file is simply a raw text file, but with Javascript syntax to allow for hierarchical structure in the data
# - Here is the same data queried above from the OpenWeatherMap API, but in JSON:
# 
# ```json
# {
#     "coord":{
#         "lon":-95.24,
#         "lat":38.97
#     },
#     "weather":[
#         {
#             "id":800,
#             "main":"Clear",
#             "description":"clear sky",
#             "icon":"01d"
#         }
#     ],
#     "base":"stations",
#     "main":{
#         "temp":297.929,
#         "pressure":996.4,
#         "humidity":27,
#         "temp_min":297.929,
#         "temp_max":297.929,
#         "sea_level":1032.03,
#         "grnd_level":996.4
#     },
#     "wind":{
#         "speed":3.04,
#         "deg":14
#     },
#     "clouds":{
#         "all":0
#     },
#     "dt":1524595410,
#     "sys":{
#         "message":0.0097,
#         "country":"US",
#         "sunrise":1524569446,
#         "sunset":1524618463
#     },
#     "id":4274277,
#     "name":"Lawrence",
#     "cod":200
# }
# ```

# # Working with APIs in Python

# - At its simplest, working with APIs involves just two components:
#     1. Pulling the data from the web server (in XML, JSON, or some other format)
#     2. Parsing the resulting text
# 
# 
# - But for APIs that receive a lot of use (such as Twitter), you'll also find special packages for pulling and processing that data 
# - We'll start with the fundamentals and then demonstrate the use of a couple of packages with APIs

# ## Requesting data
# 
# - Making an API request is not very different from pulling a regular webpage, only with most APIs we'll need to supply an API key giving us authorization to interact with the site's data
# - The key stored below as `weather_key` is my personal key that OpenWeatherMap.org gave me after I signed up for authorization to scrape their data
# 
# ```python
# openWeatherMapURL = 'http://api.openweathermap.org/data/2.5/weather'
# weather_key = '8bbb4d7931a4734804a4e6d37925c79a'
# params = {'q': 'Lawrence,us', 'appid': weather_key}
# r = requests.get(openWeatherMapURL, params)
# weather_data = json.loads(r.text)
# print(json.dumps(weather_data, indent=4))
# ```
# 
# ```json
# {
#     "coord": {
#         "lon": -95.24,
#         "lat": 38.97
#     },
#     "weather": [
#         {
#             "id": 800,
#             "main": "Clear",
#             "description": "clear sky",
#             "icon": "01d"
#         }
#     ],
#     "base": "stations",
#     "main": {
#         "temp": 296.42,
#         "pressure": 1016,
#         "humidity": 60,
#         "temp_min": 295.15,
#         "temp_max": 298.15
#     },
#     "visibility": 16093,
#     "wind": {
#         "speed": 2.6,
#         "deg": 330
#     },
#     "clouds": {
#         "all": 1
#     },
#     "dt": 1526939640,
#     "sys": {
#         "type": 1,
#         "id": 1078,
#         "message": 0.249,
#         "country": "US",
#         "sunrise": 1526900564,
#         "sunset": 1526952781
#     },
#     "id": 4274277,
#     "name": "Lawrence",
#     "cod": 200
# }
# ```

# - Here is an example of some XML data we downloaded earlier using the same method, but with the `"mode"` parameter set to `"xml"`

# In[2]:


ddir = "../../data"
weather_data = et.parse(os.path.join(ddir, "weather.xml"))
root = weather_data.getroot()
list(root)


# In[3]:


root.attrib


# - As with the HTML examples, we can walk through this file node by node

# In[4]:


weather_data.find('city').attrib


# In[5]:


weather_data.find('temperature').attrib


# In[6]:


maxTemp = weather_data.find('temperature').attrib['max']
maxTemp


# - Now let's try something more interesting
# - From the OpenWeatherMap.org site we downloaded a list of city information for use with the API
# - Then we took a random sample and saved weather data on those cities

# In[7]:


city_list = pd.read_table(os.path.join(ddir, "city_list.txt"))
city_samp = city_list.sample(10)
city_samp


# - Now we did this earlier so the random set is now different, but the following is the code we used to request and save the data locally (in the `weather_data` folder in `workingdata`)
# 
# ```python
# wdir = '../../workingdata'
# 
# for i in city_samp['id']:
# 
#     ## set parameters for API request
#     params = {'id': str(i), 'appid': weather_key}
#     
#     ## request data (in the default JSON format)
#     r = requests.get(openWeatherMapURL, params)
#     
#     ## extract city name (for the file names) and fix spaces
#     cityName = city_samp.loc[df1['id'] == i, 'nm'].iloc[0]
#     cityName = re.sub('\s+', '_', cityName)
#     
#     ## save file
#     with open(os.path.join(wdir, 'weather_data', cityName + '.json'), 'w') as outfile:
#         print(r.text, file=outfile)
# ```
# 
# - We now read in this data and plot the temperatures of the 10 randomly chosen cities

# In[8]:


## set directory
wdir = '../../workingdata'

## initialize lists
cities = []
temps = []

## read each JSON file and store info
files = os.listdir(os.path.join(wdir, 'weather_data'))
for i in files:
    with open(os.path.join(wdir, 'weather_data', i), 'r') as f:
        data = json.loads(f.read())
        cities.append(data['name'])
        temps.append(data['main']['temp'])


# In[9]:


cities


# In[10]:


temps


# In[11]:


temps = [temp*(9/5) - 459.67 for temp in temps]
temps


# In[12]:


tempDF = pd.DataFrame({'City': cities, 'Temperature': temps})
tempDF


# In[13]:


tempDF = tempDF.sort_values(by=['Temperature'], ascending=False)
tempDF


# In[14]:


get_ipython().magic('matplotlib inline')
plt.rcParams['figure.figsize'] = [12, 7]

plt.vlines(range(10), [0], tempDF['Temperature'], linewidth=4)
plt.ylim(40, 90)
plt.xticks(range(10), tempDF['City'], rotation=60, ha='right', fontsize=14);


# ## Example 1:  Twitter
# 
# - Everyone and their grandmother wants to scrape Twitter data
# - So we will show a few examples here to get you started
# 
# ### Setup: Authorization
# 
# - Twitter's API is probably one of the most heavily accessed on the Internet
# - As such, they have a bit more security and hoops to jump through to gain access
# - These are the steps I followed to gain the necessary access keys:
#     1. Sign in to your account (if you don't have Twitter you'll need to create one) at [apps.twitter.com](apps.twitter.com)
#     2. Click on "Create New App" and fill out the requested information (phone number is a requirement), but leave the "Callback URL" spot empty
#     3. Retrieve 4 strings for use in the code below:
#         - CONSUMER KEY, CONSUMER SECRET (from the "Keys and Access Tokens" tab)
#         - ACCESS TOKEN, ACCESS TOKEN SECRET (generated after clicking "Create my access token")
# - With this information we can now make requests as usual
# - First, let's try scraping tweets from a user's timeline
# 
# ### User timeline data
# 
# ```python
# from requests_oauthlib import OAuth1
# 
# ## authorization info
# key = '[CONSUMER KEY]'
# secretAPI = '[CONSUMER SECRET]'
# token = '[ACCESS TOKEN]'
# secretAcc = '[ACCESS TOKEN SECRET]'
# 
# ## URL for accessing user timelines
# timelineURL = 'https://api.twitter.com/1.1/statuses/user_timeline.json'
# 
# ## pull tweets from Liverpool and Real Madrid in anticipation of the Champions League Final
# auth1 = OAuth1(key, secretAPI, token, secretAcc)
# 
# ## scrape 10 most recent tweets
# rLFC = requests.get(timelineURL, auth=auth1, params={'screen_name': 'LFC', 'count': 10})
# rRM = requests.get(timelineURL, auth=auth1, params={'screen_name': 'realmadrid', 'count': 10})
# 
# wdir = '../../workingdata'
# with open(os.path.join(wdir, 'liverpool_tweets.json'), 'w') as outfile:
#     print(rLFC.text, file=outfile)
# with open(os.path.join(wdir, 'real-madrid_tweets.json'), 'w') as outfile:
#     print(rRM.text, file=outfile)
# ```

# In[15]:


## read from local file
## Liverpool
with open(os.path.join(wdir, 'liverpool_tweets.json'), 'r') as f:
    lfc_data = json.loads(f.read())
## Real Madrid
with open(os.path.join(wdir, 'real-madrid_tweets.json'), 'r') as f:
    rm_data = json.loads(f.read())


# - Let's check the structure a bit

# In[16]:


len(lfc_data)


# In[17]:


lfc_data[0].keys()


# In[18]:


lfc_data[0]['text']


# In[19]:


lfc_data[0]['retweet_count']


# In[20]:


lfc_data[0]['favorite_count']


# - Looks like with the retweet and favorite info we can develop some kind of a metric of fan support before the final
# - To get this data, we just need to loop over each team's data and accumulate retweets and favorites (is this called a 'like' in Twitter parlance as well?)
# - I don't use Twitter, except for data collection, so I will just treat them as the same thing

# In[21]:


## initialize
lfc_support = 0
rm_support = 0

## accumulate Liverpool support
for dat in lfc_data:
    lfc_support = lfc_support + dat['retweet_count'] + dat['favorite_count']

## accumulate Real Madrid support
for dat in rm_data:
    rm_support = rm_support + dat['retweet_count'] + dat['favorite_count']


# In[22]:


fig = plt.figure()
ax = fig.add_subplot(111)
ax.bar(1, lfc_support, color='red', edgecolor='green', lw=10, zorder = 5)
ax.bar(2, rm_support, color='yellow', edgecolor='blue', lw=10, zorder = 0)
plt.xticks([1, 2], ['Liverpool', 'Real Madrid'], fontsize=18)
plt.ylabel('Recent Tweets Favorited/Retweeted', fontsize=14)
plt.show()


# - This disparity is not a surprise to football fans, but nevertheless, **Go Reds!**
# - To do the above with the `twitter` package, we would use the following code to get LFC tweets, for example:
# 
# ```python
# import twitter
# api = twitter.Api(key, secretAPI, token, secretAcc)
# lfc_tweets = api.GetUserTimeline(screen_name='LFC', count=10)
# lfc_tweets[0].retweet_count
# ...
# ```

# ### Searching tweets
# 
# - What are people in general saying about the final?
# - Let's search for the hashtag #UCLFinal and within the recent tweets look for mentions of Liverpool or Real Madrid
# 
# ```python
# ## URL for accessing basic search
# searchURL = 'https://api.twitter.com/1.1/search/tweets.json'
# 
# r = requests.get(searchURL, auth=auth1, params={'q': '#UCLFinal', 'count': 50})
# 
# wdir = '../../workingdata'
# with open(os.path.join(wdir, 'champions-league_tweets.json'), 'w') as outfile:
#     print(r.text, file=outfile)
# ```

# In[23]:


with open(os.path.join(wdir, 'champions-league_tweets.json'), 'r') as f:
    data = json.loads(f.read())
data.keys()


# In[24]:


len(data['statuses'])


# In[25]:


data['statuses'][0]


# In[26]:


data['statuses'][0]['text']


# - We can try looping through and extracting all the instances of "Liverpool", "Real Madrid", "LFC", "RMCF", "Ronaldo", "Salah", etc.

# In[27]:


rmMentions = []
for s in data['statuses']:
    txt = s['text']
    rmMentions.append(re.findall("[Rr]eal [Mm]adrid|RMCF|[Bb]lancos|[Ww]ites|[Rr]onaldo|CR7", txt))

[m for m in rmMentions if m != []]


# In[28]:


lfcMentions = []
for s in data['statuses']:
    txt = s['text']
    lfcMentions.append(re.findall("[Ll]iverpool|[Rr]eds|Salah", txt))

[m for m in lfcMentions if m != []]


# - As you can see, Twitter scraping is no sweat, though processing the data returned from Twitter can be quite messy
# - Aren't you glad we have those regular expressions in our arsenal (no pun intended) now?
# 
# 
# **And now for something completely different!**

# ## Example 2: NASA
# 
# - NASA has a well-documented API allowing users to scrape a host of interesting types of data
# - Further, there is a specific library for interacting with the NASA API, `nasa`, that we can ultimately use to simplify the scraping process
# - But first, to better understand the API structure we will interact directly with the API
# - Here again is my personal key for the NASA API:
# 
# ```python
# nasa_key = 'JFvq5vV9kpKZDnrK5CkkPqxVFwY1pu9RyB2M5bMw'
# ```
# 
# - NASA has a bunch of different sets of data we can pull
# - Let's start with a fun favorite: downloading the Astronomy Picture of the Day (APOD) and saving it to be set as the desktop background

# ### Astronomy Picture of the Day
# 
# - First we set the base URL to work from:
# 
# ```python
# apodURL = 'https://api.nasa.gov/planetary/apod' 
# ```
# 
# - Then we pass in the key as a parameter in `request.get`
# 
# ```python
# params = {'api_key': nasa_key}
# r = requests.get(apodURL, params)
# ```
# 
# - Now the `json.loads` function converts the text string to a JSON object for ease of interaction
# 
# ```python
# apod_data = json.loads(r.text)
# ```
# - And the key bit here is we need to grab the URL and then use that to download the picture (**Note:** the image itself will not be stored in the JSON file, just the reference to its location)
# 
# ```python
# apod_data['url']
# ```
# 
# ```python
# 'https://apod.nasa.gov/apod/image/1805/TerminatorMoon_Zia_960.jpg'
# ```
# 
# - We can also grab other information, such as a description of the picture:
# 
# ```python
# apod_data['explanation']
# ```
# 
# ```python
# "Why does the right part of this image of the Moon stand out? Shadows. The terminator line -- the line between light and dark -- occurs in the featured image so that just over half the Moon's face is illuminated by sunlight. The lunar surface appears different nearer the terminator because there the Sun is nearer the horizon and therefore causes shadows to become increasingly long. These shadows make it easier for us to discern structure, giving us depth cues so that the two-dimensional image, when dominated by shadows, appears almost three-dimensional. Therefore, as the Moon fades from light to dark, shadows not only tell us the high from the low, but become noticeable for increasingly shorter structures. For example, many craters appear near the terminator because their height makes them easier to discern there. The image was taken two weeks ago when the lunar phase was waning gibbous. The next full moon, a Moon without shadows, will occur one week from today. Follow APOD on: Facebook, Google Plus, Instagram, or Twitter"
# ```
# 
# - But let's focus on the picture downloading for now
# - Now we download the file as usual:
# 
# ```
# pic = requests.get(apod_data['url'])
# ```
# 
# - Then save it (note the write argument is specified as `wb` for writing bytes):
# 
# ```
# with open("test_pic.jpg", "wb") as picFile:
#     picFile.write(pic.content)
# ```
# 
# - Let's test out how we could do this programmatically by writing a loop that will pull and save APODs from my last 3 birthdays
# - Typically we would set a directory for our pictures, but in this case I want them to be at the top level
# 
# ```python
# for year in ['2016', '2017', '2018']:
#     
#     ## set birthday dynamically
#     birthday = year + '-01-12'
#     
#     ## set params, adding in date info
#     params = {'api_key': nasa_key, 'date': birthday}
#     
#     ## pull APOD data and extract URL of picture 
#     r = requests.get(apodURL, params)
#     apod_data = json.loads(r.text)
#     picURL = apod_data['url']
#     
#     ## get image directly from picture URL
#     pic = requests.get(picURL)
#     
#     ## write out picture to file
#     with open(birthday + '.jpg', 'wb') as f:
#         f.write(pic.content)
# ```
# 
# - Below we present the three images we generated, which are just loaded directly into this markdown document (as they were saved earlier)

# <img src="./2016-01-12.jpg" alt="Jan 12 APOD, 2016">

# <img src="./2017-01-12.jpg" alt="Jan 12 APOD, 2016">

# <img src="./2018-01-12.jpg" alt="Jan 12 APOD, 2016">

# - Now on my system (Linux; I don't think this works on Windows), after the desktop background is set, it just remembers the file name in the location it was chosen from
# - So creating a program that will replace the background image each day just requires 3 components:
# 
#     1. Request the latest APOD:
#     
#     ```python
#     params = {'api_key': nasa_key}
#     r = requests.get(apodURL, params)
#     apod_data = json.loads(r.text)
#     picURL = apod_data['url']
#     pic = requests.get(picURL)
#     ```
#     
#     2. Save the result to a name of the file currently set as my desktop background
#     
#     ```python
#     with open('background.jpg', 'wb') as f:
#            f.write(pic.content)
#     ```
#     
#     3. Save the program and do whatever you have to do on your system to run `python3 setBackground.py` on a timer

# #### With the `nasa` package...
# 
# - The `nasa` package provides convenient functions for interacting with the API
# - With this package we have to supply our key in an unconventional way, however, by setting an environment variable (why the package author didn't just allow us to put it in the param list I do not know)
# 
# ```python
# os.environ['NASA_API_KEY'] = nasa_key
# ```
# 
# - What we did above could then be accomplished with the following:
# 
# ```python
# from nasa import apod
# apod_data = apod.apod('2018-01-12')
# apod_data.explanation
# ```
# 
# ```python
# "Discovered with the PanSTARRS telescope on September 7, 2016, this Comet PanSTARRS, C/2016 R2, is presently about 24 light minutes (3 AU) from the Sun, sweeping through planet Earth's skies across the background of stars in the constellation Taurus. An inbound visitor from our Solar System's distant Oort Cloud, its beautiful and complex ion tail is a remarkable shade of blue. Still relatively far from the Sun, the comet's already well-developed ion tail is very impressive. Emission from unusually abundant ionized carbon monoxide (CO+) molecules fluorescing in the increasing sunlight is largely responsible for the pretty blue tint. This color image of the blue comet is a combination of data taken from two different telescopes during the night of January 7. Located at the apex of the V-shaped Hyades star cluster in Taurus, bright star Gamma Tauri is responsible for the glow at the bottom left corner of the frame."
# ```

# ### Asteroid monitoring
# 
# - Now for my favorite part, and to leave you with a bit of paranoia to enter your weekend with:
# - NASA tracks asteroids that come near Earth, and you can access this data through their API
# - This data is stored in a separate URL so we first define that:
# 
# ```python
# neoURL = 'https://api.nasa.gov/neo/rest/v1/feed' 
# ```
# 
# - Now let's request some data. Because I'm egotistical, let's grab all the asteroids from the week leading up to my birthday this year:
# 
# ```python
# ## request data
# params = {'api_key': nasa_key, 'start_date': '2018-01-04', 'end_date': '2018-01-11'}
# r = requests.get(neoURL, params)
# 
# ## save locally for use later (and testing in class)
# wdir = '../../workingdata'
# with open(os.path.join(wdir, 'asteroid_data.json'), 'w') as outfile:
#     print(r.text, file=outfile)
# ```
# 
# 

# In[29]:


wdir = '../../workingdata'
with open(os.path.join(wdir, 'asteroid_data.json'), 'r') as f:
    ast_data = json.loads(f.read())


# In[30]:


ast_data.keys()


# In[31]:


ast_data['element_count']


# - **Whoa!** 84 near-Earth asteroids
# - Maybe I should feel lucky to have made it to 29
# - Let's see how close they got
# - The `"near_earth_objects"` key is where I believe all the important information is (you can easily browse this by pasting one of the URLs into the browser

# In[32]:


ast_data['near_earth_objects'].keys()


# In[33]:


ast_data['near_earth_objects']['2018-01-05']


# - What I want for a nice visualization is two pieces of information:
# 
#     1. The size of the asteroid
#     2. How close it came to Earth
#     
#     
# - For size, it looks like we get that in `"estimated_diameter" > "miles" > "estimated_diameter_min"`
# - For proximity, that's in `"close_approach_data" > "miss_distance" > "miles"`
# - Now we just loop through each date and store the data in our own lists

# In[34]:


## initialize lists
size = []
prox = []

neos = ast_data['near_earth_objects']

## loop over years
for year in neos.keys():
    
    ## loop over asteroids
    for i in neos[year]:
        size.append(i['estimated_diameter']['miles']['estimated_diameter_min'])
        prox.append(i['close_approach_data'][0]['miss_distance']['miles'])


# In[35]:


size[:5]


# In[36]:


prox[:5]


# In[37]:


astDF = pd.DataFrame({'Size': size, 'Proximity': prox})
astDF = astDF.astype('f')  # data should be floats for the plot
astDF = astDF.sort_values(by=['Proximity'])
closestAsteroids = astDF.head(10)
closestAsteroids


# In[38]:


1.05e4


# In[39]:


plt.rcParams['figure.figsize'] = [12, 7]

#yscale('log')
plt.hlines(range(10), [0], closestAsteroids['Proximity'], linewidth=2)
plt.scatter(closestAsteroids['Proximity'], range(10), s=[s*20000 for s in closestAsteroids['Size']])
plt.xlabel('Distance from Earth (miles)')
plt.annotate('Only 10,500 miles away, but less than 2 feet across!', xy=(20000, 0), xytext=(500000, 0.08),
             arrowprops=dict(facecolor='red', shrink=0.05, width=2, headwidth=8, edgecolor='r'), 
             color='r', fontsize=14);


# - Well that was fun (and reassuring)!
# - Now you could do the same thing as we recommended for the daily updating wallpaper (setup a script to give you daily asteroid projections), but that may get stressful
# - In any case, we recommend thinking about what tasks you may want to have automated, because that is one of the beauties of working in Python
# - There's much more to learn to become a proficient web scraper, but all we can recommend is lots of practice, and of course: **Read the Docs!**
# 
# <p style="text-align:center"><em>Happy Scraping!</em></p>
